@@ -1,6 +1,8 @@
 require 'cgi'
 require 'beaker/dsl/install_utils'
 require 'beaker/dsl/helpers'
+require 'beaker-puppet'
+require 'open3'
 require 'pp'
 require 'set'
 require 'test/unit/assertions'
@@ -10,7 +12,7 @@ require 'inifile'
 module PuppetDBExtensions
   include Test::Unit::Assertions
 
-  GitReposDir = Beaker::DSL::InstallUtils::SourcePath
+  GitReposDir = Beaker::DSL::InstallUtils::FOSSUtils::SourcePath
 
   # We include the Puppet path here so we can use the Puppet ecosystem when
   # using rake, which needs the facter gem to work - AIO ruby contains this
@@ -25,9 +27,9 @@ module PuppetDBExtensions
       [:git, :package, :pe], "install type", "PUPPETDB_INSTALL_TYPE", :git)
 
     install_mode =
-        get_option_value(options[:install_mode],
-                         [:install, :upgrade_latest, :upgrade_oldest], "install mode",
-                         "INSTALL_MODE", :install)
+        get_option_value(options[:install_type],
+                         [:install, :upgrade_latest, :upgrade_oldest], "install type",
+                         "INSTALL_TYPE", :install)
 
     validate_package_version =
         get_option_value(options[:puppetdb_validate_package_version],
@@ -264,12 +266,45 @@ module PuppetDBExtensions
     raise
   end
 
-  def get_package_version(host, version = nil)
+  def is_bionic()
+    return test_config[:os_families].has_key? 'ubuntu1804-64-1'
+  end
 
+  def oldest_supported
+    # account for special case where bionic doesn't have builds before 5.2.4
+    return is_bionic ? '5.2.4' : '5.1.1'
+  end
+
+  def get_testing_branch(version)
+    branch_name = /^((?:\d+\.)*)\d+/.match(version)[1] + 'x'
+    if branch_name.chars.first.to_i > 5
+      branch_name = 'master'
+    end
+    return branch_name
+  end
+
+  def get_latest_released(version)
+    cloned = system('git clone https://github.com/puppetlabs/puppetdb.git')
+    if cloned.nil?
+      raise 'error cloning puppetdb repo'
+    end
+    branch_name = get_testing_branch(version)
+    stdout, status = Open3.capture2('git', '-C', 'puppetdb', 'describe', '--tags', '--abbrev=0', "origin/#{branch_name}")
+    if status.exitstatus != 0
+      raise "error getting most recent tagged release. status: #{status}"
+    end
+    return stdout.delete!("\n")
+  end
+
+  def get_package_version(host, version = nil)
     if version == 'latest'
       return 'latest'
     elsif version.nil?
       version = PuppetDBExtensions.config[:package_build_version].to_s
+      # If no version was defined, default to latest.
+      if version == ''
+        return 'latest'
+      end
     end
 
     # version can look like:
@@ -295,8 +330,14 @@ module PuppetDBExtensions
     elsif host['platform'].include?('fedora')
       version_tag = host['platform'].match(/^fedora-(\d+)/)[1]
       "#{version}.fc#{version_tag}"
-    elsif host['platform'].include?('ubuntu') or host['platform'].include?('debian')
-      "#{version}puppetlabs1"
+    elsif host['platform'].include?('ubuntu-16.04')
+      "#{version}xenial"
+    elsif host['platform'].include?('ubuntu-18.04')
+      "#{version}bionic"
+    elsif host['platform'].include?('debian-8')
+      "#{version}jessie"
+    elsif host['platform'].include?('debian-9')
+      "#{version}stretch"
     else
       raise ArgumentError, "Unsupported platform: '#{host['platform']}'"
     end
@@ -309,6 +350,7 @@ module PuppetDBExtensions
     }
     class { 'puppetdb::server':
       manage_firewall => false,
+      disable_update_checking => true,
     }
     #{postgres_manifest}
     Postgresql::Server::Db['puppetdb'] -> Class['puppetdb::server']
@@ -335,7 +377,7 @@ module PuppetDBExtensions
       expected_version = get_package_version(host)
 
       Beaker::Log.notify "Expecting package version: '#{expected_version}', actual version: '#{installed_version}'"
-      if installed_version != expected_version
+      if installed_version != expected_version and expected_version != 'latest'
         raise RuntimeError, "Installed version '#{installed_version}' did not match expected version '#{expected_version}'"
       end
     end
@@ -383,12 +425,22 @@ module PuppetDBExtensions
     CGI.escape(Time.rfc2822(result.stdout).iso8601)
   end
 
+  def enable_https_apt_sources(host)
+    manifest = <<-EOS
+      if ($facts['osfamily'] == 'Debian') {
+        package { 'apt-transport-https': ensure => installed }
+      }
+    EOS
+
+    apply_manifest_on(host, manifest)
+  end
+
   def postgres_manifest
     manifest = <<-EOS
       # get the pg server up and running
       class { 'postgresql::globals':
           manage_package_repo => true,
-          version             => '9.4',
+          version             => '9.6',
       }->
       class { '::postgresql::server':
         ip_mask_allow_all_users => '0.0.0.0/0',
@@ -717,9 +769,106 @@ EOS
   # End Object diff functions
   ##############################################################################
 
+  def initialize_repo_on_host(host, os, nightly)
+    case os
+    when :debian
+
+      # For openjdk8
+      if host['platform'].version == '8'
+        create_remote_file(host,
+                          "/etc/apt/sources.list.d/jessie-backports.list",
+                          "deb http://httpredir.debian.org/debian jessie-backports main")
+      end
+
+      if options[:type] == 'aio' then
+        if nightly
+          ## puppet6 repos
+          on host, "curl -O http://nightlies.puppetlabs.com/apt/puppet6-nightly-release-$(lsb_release -sc).deb"
+          on host, "dpkg -i puppet6-nightly-release-$(lsb_release -sc).deb"
+
+        else
+          on host, "curl -O http://apt.puppetlabs.com/puppet-release-$(lsb_release -sc).deb"
+          on host, "dpkg -i puppet-release-$(lsb_release -sc).deb"
+        end
+      else
+        on host, "curl -O http://apt.puppetlabs.com/puppetlabs-release-$(lsb_release -sc).deb"
+        on host, "dpkg -i puppetlabs-release-$(lsb_release -sc).deb"
+      end
+      on host, "apt-get update"
+      on host, "apt-get install debian-archive-keyring"
+    when :redhat
+      if options[:type] == 'aio' then
+        /^(el|centos)-(\d+)-(.+)$/.match(host.platform)
+        variant = ($1 == 'centos') ? 'el' : $1
+        version = $2
+        arch = $3
+
+        if nightly
+          ## puppet6 repos
+          on host, "curl -O http://yum.puppetlabs.com/puppet6-nightly/puppet6-nightly-release-#{variant}-#{version}.noarch.rpm"
+          on host, "rpm -i puppet6-nightly-release-#{variant}-#{version}.noarch.rpm"
+
+        else
+          on host, "curl -O http://yum.puppetlabs.com/puppet/puppet-release-#{variant}-#{version}.noarch.rpm"
+          on host, "rpm -i puppet-release-#{variant}-#{version}.noarch.rpm"
+        end
+      else
+        on host, "yum clean all -y"
+        on host, "yum upgrade -y"
+        create_remote_file host, '/etc/yum.repos.d/puppetlabs-dependencies.repo', <<-REPO.gsub(' '*8, '')
+  [puppetlabs-dependencies]
+  name=Puppet Labs Dependencies - $basearch
+  baseurl=http://yum.puppetlabs.com/el/$releasever/dependencies/$basearch
+  gpgkey=http://yum.puppetlabs.com/RPM-GPG-KEY-puppetlabs
+  enabled=1
+  gpgcheck=1
+        REPO
+
+        create_remote_file host, '/etc/yum.repos.d/puppetlabs-products.repo', <<-REPO.gsub(' '*8, '')
+  [puppetlabs-products]
+  name=Puppet Labs Products - $basearch
+  baseurl=http://yum.puppetlabs.com/el/$releasever/products/$basearch
+  gpgkey=http://yum.puppetlabs.com/RPM-GPG-KEY-puppetlabs
+  enabled=1
+  gpgcheck=1
+        REPO
+
+        create_remote_file host, '/etc/yum.repos.d/epel.repo', <<-REPO
+  [epel]
+  name=Extra Packages for Enterprise Linux $releasever - $basearch
+  baseurl=http://download.fedoraproject.org/pub/epel/$releasever/$basearch
+  mirrorlist=https://mirrors.fedoraproject.org/metalink?repo=epel-$releasever&arch=$basearch
+  failovermethod=priority
+  enabled=1
+  gpgcheck=0
+        REPO
+      end
+    when :fedora
+      create_remote_file host, '/etc/yum.repos.d/puppetlabs-dependencies.repo', <<-REPO.gsub(' '*8, '')
+  [puppetlabs-dependencies]
+  name=Puppet Labs Dependencies - $basearch
+  baseurl=http://yum.puppetlabs.com/fedora/f$releasever/dependencies/$basearch
+  gpgkey=http://yum.puppetlabs.com/RPM-GPG-KEY-puppetlabs
+  enabled=1
+  gpgcheck=1
+      REPO
+
+      create_remote_file host, '/etc/yum.repos.d/puppetlabs-products.repo', <<-REPO.gsub(' '*8, '')
+  [puppetlabs-products]
+  name=Puppet Labs Products - $basearch
+  baseurl=http://yum.puppetlabs.com/fedora/f$releasever/products/$basearch
+  gpgkey=http://yum.puppetlabs.com/RPM-GPG-KEY-puppetlabs
+  enabled=1
+  gpgcheck=1
+      REPO
+    else
+      raise ArgumentError, "Unsupported OS '#{os}'"
+    end
+  end
+
   def install_puppet_from_package
     hosts.each do |host|
-      install_package(host, 'puppet-agent')
+      install_puppet_agent_on(host, {:puppet_collection => "puppet6"})
       on( host, puppet('resource', 'host', 'updates.puppetlabs.com', 'ensure=present', "ip=127.0.0.1") )
       install_package(host, 'puppetserver')
     end
