@@ -110,6 +110,14 @@ describe Puppet::Resource::Catalog::Puppetdb do
       end
     end
 
+    describe "#add_job_id_if_missing" do
+      it "should add the job_id key" do
+        result = subject.add_job_id_if_missing(catalog_data_hash, nil)
+        result.has_key?('job_id').should be_truthy
+        result['job_id'].should be_nil
+      end
+    end
+
     describe "#add_producer" do
       it "should add the producer key" do
         result = subject.add_producer(catalog_data_hash, "foo")
@@ -129,7 +137,7 @@ describe Puppet::Resource::Catalog::Puppetdb do
 
       it "should leave an existing parameters hash alone" do
         msg = "with up so floating many bells down"
-        resource[:message] = msg
+        resource['message'] = msg
 
         result = subject.add_parameters_if_missing(catalog_data_hash)
         resource = result['resources'].find do |res|
@@ -137,7 +145,7 @@ describe Puppet::Resource::Catalog::Puppetdb do
         end
 
         resource.should_not be_nil
-        resource['parameters'].should == {:message => msg}
+        resource['parameters'].should == {'message' => msg}
       end
     end
 
@@ -154,7 +162,7 @@ describe Puppet::Resource::Catalog::Puppetdb do
         end
 
         resource.should_not be_nil
-        resource['parameters'][:alias].should include(name)
+        resource['parameters']['alias'].should include(name)
       end
 
       context "with resource types that provide #title_patterns" do
@@ -195,8 +203,8 @@ describe Puppet::Resource::Catalog::Puppetdb do
             #  this test should cover other resource types that fall into
             #  this category as well.
             resource.should_not be_nil
-            resource['parameters'][:alias].should_not be_nil
-            resource['parameters'][:alias].should include('/tmp/foo')
+            resource['parameters']['alias'].should_not be_nil
+            resource['parameters']['alias'].should include('/tmp/foo')
           end
         end
       end
@@ -210,7 +218,7 @@ describe Puppet::Resource::Catalog::Puppetdb do
         end
 
         resource.should_not be_nil
-        resource['parameters'][:alias].should be_nil
+        resource['parameters']['alias'].should be_nil
       end
 
       describe "for resources with composite namevars" do
@@ -229,7 +237,7 @@ describe Puppet::Resource::Catalog::Puppetdb do
           end
 
           resource.should_not be_nil
-          resource['parameters'][:alias].should be_nil
+          resource['parameters']['alias'].should be_nil
         end
       end
 
@@ -247,7 +255,7 @@ describe Puppet::Resource::Catalog::Puppetdb do
           end
 
           resource.should_not be_nil
-          resource['parameters'][:alias].should == ['something awesome']
+          resource['parameters']['alias'].should == ['something awesome']
         end
       end
     end
@@ -269,9 +277,9 @@ describe Puppet::Resource::Catalog::Puppetdb do
         end
 
         resource.should_not be_nil
-        resource['parameters'][:command].should == '/bin/true'
-        resource['parameters'][:path].should == ['/foo/goo', '/foo/bar']
-        resource['parameters'][:audit].should == 'path'
+        resource['parameters']['command'].should == '/bin/true'
+        resource['parameters']['path'].should == ['/foo/goo', '/foo/bar']
+        resource['parameters']['audit'].should == 'path'
       end
 
       it "should sort unordered metaparams with array values" do
@@ -283,8 +291,8 @@ describe Puppet::Resource::Catalog::Puppetdb do
         end
 
         resource.should_not be_nil
-        resource['parameters'][:audit].should == 'path'
-        resource['parameters'][:tag].should == ['a', 'b', 'c']
+        resource['parameters']['audit'].should == 'path'
+        resource['parameters']['tag'].should == ['a', 'b', 'c']
       end
     end
 
@@ -543,15 +551,17 @@ describe Puppet::Resource::Catalog::Puppetdb do
         end
       end
 
-      it "should add edges even if the other end is an alias" do
-        other_resource = Puppet::Resource.new(:notify, 'noone', :parameters => {:alias => 'another_thing'})
-        resource[:require] = 'Notify[another_thing]'
-        Puppet[:code] = [resource, other_resource].map(&:to_manifest).join
+      it "should produce an edge when referencing an aliased resource that only has a single namevar" do
+        Puppet[:code] = <<-MANIFEST
+        notify { 'noone':
+          alias => 'another_thing'
+        }
+        notify { 'anyone':
+          require => Notify['another_thing']
+        }
+        MANIFEST
 
-        hash = catalog.to_data_hash
-        subject.add_parameters_if_missing(hash)
-        subject.add_namevar_aliases(hash, catalog)
-        result = subject.synthesize_edges(hash, catalog)
+        result = subject.munge_catalog(catalog, Time.now.utc)
 
         edge = {'source' => {'type' => 'Notify', 'title' => 'noone'},
                 'target' => {'type' => 'Notify', 'title' => 'anyone'},
@@ -560,11 +570,31 @@ describe Puppet::Resource::Catalog::Puppetdb do
         result['edges'].should include(edge)
       end
 
+      it "should produce an edge when referencing an aliased resource that supports composite namevars" do
+        Puppet[:code] = <<-MANIFEST
+        package { 'foo':
+          ensure => present,
+          alias => 'bar'
+        }
+        notify { 'hello':
+          require => Package['bar']
+        }
+        MANIFEST
+
+        result = subject.munge_catalog(catalog, Time.now.utc)
+
+        edge = {'source' => {'type' => 'Package', 'title' => 'foo'},
+                'target' => {'type' => 'Notify', 'title' => 'hello'},
+                'relationship' => 'required-by'}
+
+        result['edges'].should include(edge)
+      end
+
       it "should produce a reasonable error message for a missing 'before' relationship" do
 
         error = Gem::Version.new(Puppet.version) < Gem::Version.new("5.0.0") ?
           "Invalid relationship: Notify[anyone] { before => Notify[non-existent] }, because Notify[non-existent] doesn't seem to be in the catalog" :
-          "Could not find resource 'Notify[non-existent]' in parameter 'before' at line 2 on node node"
+          "Could not find resource 'Notify[non-existent]' in parameter 'before' (line: 2) on node node"
 
         expect {
           resource[:before] = 'Notify[non-existent]'
@@ -576,7 +606,7 @@ describe Puppet::Resource::Catalog::Puppetdb do
       it "should produce a reasonable error message for a missing 'required-by' relationship" do
         error = Gem::Version.new(Puppet.version) < Gem::Version.new("5.0.0") ?
           "Invalid relationship: Notify[anyone] { require => Notify[non-existent] }, because Notify[non-existent] doesn't seem to be in the catalog" :
-          "Could not find resource 'Notify[non-existent]' in parameter 'require' at line 2 on node node"
+          "Could not find resource 'Notify[non-existent]' in parameter 'require' (line: 2) on node node"
 
         expect {
           resource[:require] = 'Notify[non-existent]'
@@ -588,7 +618,7 @@ describe Puppet::Resource::Catalog::Puppetdb do
       it "should produce a reasonable error message for a missing 'notifies' relationship" do
         error = Gem::Version.new(Puppet.version) < Gem::Version.new("5.0.0") ?
           "Invalid relationship: Notify[anyone] { notify => Notify[non-existent] }, because Notify[non-existent] doesn't seem to be in the catalog" :
-          "Could not find resource 'Notify[non-existent]' in parameter 'notify' at line 2 on node node"
+          "Could not find resource 'Notify[non-existent]' in parameter 'notify' (line: 2) on node node"
         expect {
           resource[:notify] = 'Notify[non-existent]'
           hash = subject.add_parameters_if_missing(catalog_data_hash)
@@ -599,7 +629,7 @@ describe Puppet::Resource::Catalog::Puppetdb do
       it "should produce a reasonable error message for a missing 'subscription-of' relationship" do
         error = Gem::Version.new(Puppet.version) < Gem::Version.new("5.0.0") ?
           "Invalid relationship: Notify[anyone] { subscribe => Notify[non-existent] }, because Notify[non-existent] doesn't seem to be in the catalog" :
-          "Could not find resource 'Notify[non-existent]' in parameter 'subscribe' at line 2 on node node"
+          "Could not find resource 'Notify[non-existent]' in parameter 'subscribe' (line: 2) on node node"
 
         expect {
           resource[:subscribe] = 'Notify[non-existent]'
@@ -611,7 +641,7 @@ describe Puppet::Resource::Catalog::Puppetdb do
       it "should produce a reasonable error message for an invalid resourceref" do
         error = Gem::Version.new(Puppet.version) < Gem::Version.new("5.0.0") ?
           "Invalid relationship: Notify[anyone] { subscribe => Foobar::baz[name] }, because Foobar::baz[name] doesn't seem to be in the correct format. Resource references should be formatted as: Classname['title'] or Modulename::Classname['title'] (take careful note of the capitalization)." :
-          "Could not find resource 'Foobar::baz[name]' in parameter 'subscribe' at line 2 on node node"
+          "Could not find resource 'Foobar::baz[name]' in parameter 'subscribe' (line: 2) on node node"
 
         expect {
           resource[:subscribe] = 'Foobar::baz[name]'
@@ -699,7 +729,36 @@ describe Puppet::Resource::Catalog::Puppetdb do
 
         result.keys.should =~ ['certname', 'version', 'edges', 'resources',
           'transaction_uuid', 'environment', 'producer_timestamp', "code_id",
-          "catalog_uuid", "producer"]
+          "job_id", "catalog_uuid", "producer"]
+      end
+
+      context 'when dealing with TagSets' do
+        let(:resource_hashes) do
+          {
+            'resources' => [
+              {'type' => 'Stage',
+              'title' => :main,
+              'tags' => Puppet::Util::TagSet.new(['stage']),
+              'exported' => false,
+              'parameters' => {:name => 'main'}},
+             {'type' => 'Class',
+              'title' => 'Settings',
+              'tags' => Puppet::Util::TagSet.new(['class', 'settings']),
+              'exported' => false},
+             {'type' => 'Class',
+              'title' => :main,
+              'tags' => Puppet::Util::TagSet.new(['class']),
+              'exported' => false,
+              'parameters' => {:name => 'main'}}
+            ]
+          }
+        end
+
+        it "should convert TagSet for resource['tags']" do
+          subject.hashify_tags(resource_hashes)['resources'].each do |resource|
+            resource['tags'].should be_an Array
+          end
+        end
       end
     end
 
@@ -708,31 +767,32 @@ describe Puppet::Resource::Catalog::Puppetdb do
       let(:has_secret?) {
         lambda { |r| r['parameters'] && r['parameters'][:message] == secret}
       }
-      let(:input) {{'tags' => Puppet::Util::TagSet.new(['settings']),
+      let(:input) {{'tags' => ['settings'],
                     'name' => 'my_agent',
                     'version' => 1490991352,
                     'code_id' => nil,
+                    'job_id' => nil,
                     'catalog_uuid' => 'aa4759d3-f1f1-47a0-925c-a4acd0c1b4ed',
                     'catalog_format' => 1,
                     'environment' => 'production',
                     'resources' =>
                     [{'type' => 'Stage',
                       'title' => :main,
-                      'tags' => Puppet::Util::TagSet.new(['stage']),
+                      'tags' => ['stage'],
                       'exported' => false,
                       'parameters' => {:name => 'main'}},
                      {'type' => 'Class',
                       'title' => 'Settings',
-                      'tags' => Puppet::Util::TagSet.new(['class', 'settings']),
+                      'tags' => ['class', 'settings'],
                       'exported' => false},
                      {'type' => 'Class',
                       'title' => :main,
-                      'tags' => Puppet::Util::TagSet.new(['class']),
+                      'tags' => ['class'],
                       'exported' => false,
                       'parameters' => {:name => 'main'}},
                      {'type' => 'Notify',
                       'title' => 'hi',
-                      'tags' => Puppet::Util::TagSet.new(['notify', 'hi', 'class']),
+                      'tags' => ['notify', 'hi', 'class'],
                       'file' =>  'site.pp',
                       'line' => 1,
                       'exported' => false,
