@@ -50,6 +50,8 @@
             [overtone.at-at :refer [mk-pool every interspaced stop-and-reset-pool!]]
             [puppetlabs.kitchensink.core :as kitchensink]
             [puppetlabs.puppetdb.cheshire :as json]
+            [puppetlabs.puppetdb.cli.tk-util :refer [run-tk-cli-cmd]]
+            [puppetlabs.puppetdb.cli.util :refer [exit]]
             [puppetlabs.puppetdb.command.constants :refer [command-names]]
             [puppetlabs.puppetdb.command.dlo :as dlo]
             [puppetlabs.puppetdb.config :as conf]
@@ -78,17 +80,44 @@
             [puppetlabs.puppetdb.command :as cmd]
             [puppetlabs.puppetdb.queue :as queue]
             [puppetlabs.i18n.core :refer [trs tru]])
-  (:import [javax.jms ExceptionListener]
-           [java.util.concurrent.locks ReentrantLock]
+  (:import [java.util.concurrent.locks ReentrantLock]
            [org.joda.time Period]))
 
-(def cli-description "Main PuppetDB daemon")
 (def database-metrics-registry (get-in metrics/metrics-registries [:database :registry]))
 
-;; ## Wiring
-;;
-;; The following functions setup interaction between the main
-;; PuppetDB components.
+(def clean-options
+  #{"expire_nodes" "purge_nodes" "purge_reports" "gc_packages" "other"})
+
+(def purge-nodes-opts-schema {:batch_limit s/Int})
+
+(def clean-request-schema
+  ;; i.e. a possibly empty collection of elements, each of which must
+  ;; either be a string from clean-options or a purge_nodes command
+  ;; with an options map.
+  [(s/if string?
+     (apply s/enum clean-options)
+     [(s/one (s/eq "purge_nodes") "command")
+      (s/one purge-nodes-opts-schema "options")])])
+
+(defn reduce-clean-request [request]
+  "Converts the incoming vector of requests to a map of requests to
+  their options, where the last one of each kind wins.
+  e.g. [\"purge_nodes\" \"purge_reports\" {\"purge_nodes\" {...}}]
+  becomes #{{\"purge_reports\" true} {\"purge_nodes\" {...}}}."
+  (into {} (map (fn [x]
+                  (if (string? x)
+                    [x true]
+                    x))
+            request)))
+
+(defn reduced-clean-request->status [request]
+  (->> request keys sort
+       (map {"expire_nodes" "expiring-nodes"
+             "purge_nodes" "purging-nodes"
+             "purge_reports" "purging-reports"
+             "gc_packages" "package-gc"
+             "other" "other"})
+       (str/join " ")))
 
 (defn auto-expire-nodes!
   "Expire nodes which haven't had any activity (catalog/fact submission)
@@ -106,18 +135,22 @@
     (catch Exception e
       (log/error e (trs "Error while deactivating stale nodes")))))
 
-(defn purge-nodes!
-  "Delete nodes which have been *deactivated or expired* longer than
-  `node-purge-ttl`."
-  [node-purge-ttl db]
-  {:pre [(map? db)
-         (period? node-purge-ttl)]}
+(defn-validated purge-nodes!
+  "Deletes nodes which have been *deactivated or expired* longer than
+  node-purge-ttl.  Deletes at most batch_limit nodes if opts is a map,
+  all relevant nodes otherwise."
+  [node-purge-ttl :- (s/pred period?)
+   opts :- (s/if map? purge-nodes-opts-schema (s/eq true))
+   db :- (s/pred map?)]
   (try
     (kitchensink/demarcate
      (format "purge deactivated and expired nodes (threshold: %s)"
              (format-period node-purge-ttl))
-     (jdbc/with-transacted-connection db
-       (scf-store/purge-deactivated-and-expired-nodes! (ago node-purge-ttl))))
+     (let [horizon (ago node-purge-ttl)]
+       (jdbc/with-transacted-connection db
+         (if-let [limit (and (map? opts) (:batch_limit opts))]
+           (scf-store/purge-deactivated-and-expired-nodes! horizon limit)
+           (scf-store/purge-deactivated-and-expired-nodes! horizon)))))
     (catch Exception e
       (log/error e (trs "Error while purging deactivated and expired nodes")))))
 
@@ -135,15 +168,19 @@
     (catch Exception e
       (log/error e (trs "Error while sweeping reports")))))
 
-(defn gc-packages! [ttl db]
+(defn gc-packages! [db]
   {:pre [(map? db)]}
   (try
-    (kitchensink/demarcate (trs "package gc")
+    (kitchensink/demarcate (trs "gc packages")
       (jdbc/with-transacted-connection db
         (scf-store/delete-unassociated-packages!)))
     (catch Exception e
       (log/error e (trs "Error while running package gc")))))
 
+(defn gc-resource-events!
+  []
+  (trs "Error while sweeping resource events"))
+
 (defn garbage-collect!
   "Perform garbage collection on `db`, which means deleting any orphaned data.
   This basically just wraps the corresponding scf.storage function with some
@@ -157,16 +194,6 @@
     (catch Exception e
       (log/error e (trs "Error during garbage collection")))))
 
-(def clean-options #{"expire_nodes" "purge_nodes" "purge_reports" "package_gc" "other"})
-
-(defn clean-options->status [options]
-  (str/join " " (map {"expire_nodes" "expiring-nodes"
-                      "purge_nodes" "purging-nodes"
-                      "purge_reports" "purging-reports"
-                      "gc_packages" "package-gc"
-                      "other" "other"}
-                     (sort options))))
-
 (def admin-metrics-registry
   (get-in metrics/metrics-registries [:admin :registry]))
 
@@ -189,9 +216,6 @@
    :package-gc-time (timer admin-metrics-registry "package-gc-time")
    :other-clean-time (timer admin-metrics-registry ["other-clean-time"])})
 
-(def clean-request-schema
-  [(apply s/enum clean-options)])
-
 (defn- clear-clean-status!
   "Clears the clean status (as a separate function to support tests)."
   []
@@ -210,17 +234,19 @@
    request :- clean-request-schema]
   (when-not (.isHeldByCurrentThread lock)
     (throw (IllegalStateException. (tru "cleanup lock is not already held"))))
-  (let [request (if (empty? request) clean-options (set request))
-        status (clean-options->status request)]
+  (let [request (reduce-clean-request (if (empty? request)
+                                        clean-options  ; clean everything
+                                        request))
+        status (reduced-clean-request->status request)]
     (try
       (reset! clean-status status)
       (when (request "expire_nodes")
         (time! (:node-expiration-time admin-metrics)
                (auto-expire-nodes! node-ttl db))
         (counters/inc! (:node-expirations admin-metrics)))
-      (when (request "purge_nodes")
+      (when-let [opts (request "purge_nodes")]
         (time! (:node-purge-time admin-metrics)
-               (purge-nodes! node-purge-ttl db))
+               (purge-nodes! node-purge-ttl opts db))
         (counters/inc! (:node-purges admin-metrics)))
       (when (request "purge_reports")
         (time! (:report-purge-time admin-metrics)
@@ -265,18 +291,25 @@
              :desc "A reoccuring job to checkin the PuppetDB version"))
     (log/debug (trs "Skipping update check on Puppet Enterprise"))))
 
+(def stop-gc-wait-ms (constantly 5000))
+
 (defn stop-puppetdb
   "Shuts down PuppetDB, releasing resources when possible.  If this is
   not a normal shutdown, emergency? must be set, which currently just
   produces a fatal level level log message, instead of info."
-  [context]
+  [{:keys [stop-status] :as context}]
   (log/info (trs "Shutdown request received; puppetdb exiting."))
+  (when-let [pool (:job-pool context)]
+    (stop-and-reset-pool! pool))
+  ;; Wait up to ~5s for https://bugs.openjdk.java.net/browse/JDK-8176254
+  (swap! stop-status #(conj % :stopping))
+  (if (utils/wait-for-ref-state stop-status (stop-gc-wait-ms) #(= % #{:stopping}))
+    (log/info (trs "Periodic activities halted"))
+    (log/info (trs "Forcibly terminating periodic activities")))
   (when-let [ds (get-in context [:shared-globals :scf-write-db :datasource])]
     (.close ds))
   (when-let [ds (get-in context [:shared-globals :scf-read-db :datasource])]
     (.close ds))
-  (when-let [pool (:job-pool context)]
-    (stop-and-reset-pool! pool))
   (when-let [command-loader (:command-loader context)]
     (future-cancel command-loader))
   context)
@@ -288,11 +321,13 @@
   the current database is not supported."
   [db-conn-pool config]
   (jdbc/with-db-connection db-conn-pool
-    (let [current (:version @sutils/db-metadata)]
-      (when (neg? (compare current scf-store/oldest-supported-db))
+    (let [current (:version @sutils/db-metadata)
+          oldest (get-in config [:database :min-required-version]
+                           scf-store/oldest-supported-db)]
+      (when (neg? (compare current oldest))
         (throw+ {:type ::unsupported-database
                  :current current
-                 :oldest scf-store/oldest-supported-db}))
+                 :oldest oldest}))
       @sutils/db-metadata
       (let [migrated? (migrate! db-conn-pool)]
         (indexes! config)
@@ -325,96 +360,120 @@
       result
       (recur db-spec))))
 
+(defn db-config->clean-request
+  [config]
+  (let [seconds-pos? (comp pos? to-seconds)]
+    (filter identity
+            [(when (some-> (:node-ttl config) seconds-pos?)
+               "expire_nodes")
+             (when (some-> (:node-purge-ttl config) seconds-pos?)
+               (let [limit (:node-purge-gc-batch-limit config)]
+                 (if (zero? limit)
+                            "purge_nodes"
+                            ["purge_nodes" {:batch_limit limit}])))
+             (when (some-> (:report-ttl config) seconds-pos?)
+               "purge_reports")
+             "gc_packages"
+             "other"])))
+
+(defn collect-garbage
+  [db clean-lock config clean-request]
+  (.lock clean-lock)
+  (try
+    (try
+      (clean-up db clean-lock config clean-request)
+      (catch Exception ex
+        (log/error ex)))
+    (finally
+      (.unlock clean-lock))))
+
+(defn coordinate-gc-with-shutdown
+  [write-db clean-lock database request stop-status]
+  ;; This function assumes it will never be called concurrently with
+  ;; itself, so that it's ok to just conj/disj below, instead of
+  ;; tracking an atomic count or similar.
+  (let [update-if-not-stopping #(if (:stopping %)
+                                  %
+                                  (conj % :collecting-garbage))]
+    (try
+      (when (:collecting-garbage (swap! stop-status update-if-not-stopping))
+        (collect-garbage write-db clean-lock database request))
+      (finally
+        (swap! stop-status disj :collecting-garbage)))))
+
 (defn start-puppetdb
   "Throws {:type ::unsupported-database :current version :oldest version} if
   the current database is not supported."
-  [context config service get-registered-endpoints]
+  [context config service get-registered-endpoints upgrade-and-exit?]
   {:pre [(map? context)
          (map? config)]
    :post [(map? %)]}
-  (let [{:keys [developer jetty
-                database read-database
-                puppetdb command-processing]} config
-        {:keys [pretty-print max-enqueued]} developer
-        {:keys [gc-interval]} database
-        {:keys [disable-update-checking]} puppetdb
 
-        write-db (jdbc/pooled-datasource (assoc database :pool-name "PDBWritePool")
-                                         database-metrics-registry)
-        read-db (jdbc/pooled-datasource (assoc read-database :read-only? true :pool-name "PDBReadPool")
-                                        database-metrics-registry)]
+  (let [{:keys [database developer read-database]} config
+        write-db (-> (assoc database :pool-name "PDBWritePool")
+                     (jdbc/pooled-datasource database-metrics-registry))
+        read-db (-> (assoc read-database
+                           :pool-name "PDBReadPool"
+                           :read-only? true)
+                    (jdbc/pooled-datasource database-metrics-registry))
+        globals {:scf-read-db read-db
+                 :scf-write-db write-db
+                 :pretty-print (:pretty-print developer)}]
 
     (when-let [v (version/version)]
       (log/info (trs "PuppetDB version {0}" v)))
-
     (init-with-db database config)
 
-    (let [population-registry (get-in metrics/metrics-registries [:population :registry])]
-      (pop/initialize-population-metrics! population-registry read-db))
+    (if upgrade-and-exit?
+      (assoc context :shared-globals globals)
+      (do
+        (let [population-registry (get-in metrics/metrics-registries [:population :registry])]
+          (pop/initialize-population-metrics! population-registry read-db))
 
-    ;; Error handling here?
-    (let [stockdir (conf/stockpile-dir config)
-          command-chan (async/chan
-                         (queue/sorted-command-buffer
-                          max-enqueued
-                          #(cmd/update-counter! :invalidated %1 %2 inc!)))
-          [q load-messages] (queue/create-or-open-stockpile (conf/stockpile-dir config))
-          globals {:scf-read-db read-db
-                   :scf-write-db write-db
-                   :pretty-print pretty-print
-                   :q q
-                   :dlo (dlo/initialize (get-path stockdir "discard")
-                                        (get-in metrics-registries
-                                                [:dlo :registry]))
-                   :command-chan command-chan}
-          clean-lock (ReentrantLock.)
-          command-loader (when load-messages
-                           (future
-                             (load-messages command-chan cmd/inc-cmd-depth)))]
+        ;; Error handling here?
+        (let [stockdir (conf/stockpile-dir config)
+              command-chan (async/chan
+                            (queue/sorted-command-buffer
+                             (:max-enqueued developer)
+                             #(cmd/update-counter! :invalidated %1 %2 inc!)))
+              [q load-messages] (queue/create-or-open-stockpile (conf/stockpile-dir config))
+              globals (assoc globals
+                             :q q
+                             :dlo (dlo/initialize (get-path stockdir "discard")
+                                                  (get-in metrics-registries
+                                                          [:dlo :registry]))
+                             :command-chan command-chan)
+              clean-lock (ReentrantLock.)
+              command-loader (when load-messages
+                               (future
+                                 (load-messages command-chan cmd/inc-cmd-depth)))]
 
-      ;; Pretty much this helper just knows our job-pool and gc-interval
-      (let [job-pool (mk-pool)
-            gc-interval-millis (to-millis gc-interval)]
-        (when-not disable-update-checking
-          (maybe-check-for-updates config read-db job-pool))
-        (when (pos? gc-interval-millis)
-          (let [seconds-pos? (comp pos? to-seconds)
-                what (filter identity
-                             [(when-let [node-ttl (:node-ttl database)]
-                                (when (seconds-pos? node-ttl)
-                                  "expire_nodes"))
-                              (when-let [node-purge-ttl (:node-purge-ttl database)]
-                                (when (seconds-pos? node-purge-ttl)
-                                  "purge_nodes"))
-                              (when-let [report-ttl (:report-ttl database)]
-                                (when (seconds-pos? report-ttl)
-                                  "purge_reports"))
-                              "other"])]
-            (interspaced gc-interval-millis
-                         (fn []
-                           (.lock clean-lock)
-                           (try
-                             (try
-                               (clean-up write-db clean-lock database what)
-                               (catch Exception ex
-                                 (log/error ex)))
-                             (finally
-                               (.unlock clean-lock))))
-                         job-pool)))
-        (assoc context
-               :job-pool job-pool
-               :shared-globals globals
-               :clean-lock clean-lock
-               :command-loader command-loader)))))
+          ;; Pretty much this helper just knows our job-pool and gc-interval
+          (let [job-pool (mk-pool)
+                gc-interval-millis (to-millis (:gc-interval database))]
+            (when-not (get-in config [:puppetdb :disable-update-checking])
+              (maybe-check-for-updates config read-db job-pool))
+            (when (pos? gc-interval-millis)
+              (let [request (db-config->clean-request database)]
+                (interspaced gc-interval-millis
+                             #(coordinate-gc-with-shutdown write-db clean-lock
+                                                           database request
+                                                           (:stop-status context))
+                             job-pool)))
+            (assoc context
+                   :job-pool job-pool
+                   :shared-globals globals
+                   :clean-lock clean-lock
+                   :command-loader command-loader)))))))
 
 (defn db-unsupported-msg
-  "Returns a message describing which databases are supported."
-  [current oldest]
-  (trs "PostgreSQL {0}.{1} is no longer supported.  Please upgrade to {2}.{3}."
-       (first current)
-       (second current)
-       (first oldest)
-       (second oldest)))
+    "Returns a message describing which databases are supported."
+    [current oldest]
+    (trs "PostgreSQL {0}.{1} is no longer supported.  Please upgrade to {2}.{3}."
+         (first current)
+         (second current)
+         (first oldest)
+         (second oldest)))
 
 (defprotocol PuppetDBServer
   (shared-globals [this])
@@ -439,20 +498,32 @@
   PuppetDBServer
   [[:DefaultedConfig get-config]
    [:WebroutingService add-ring-handler get-registered-endpoints]
-   [:ShutdownService shutdown-on-error]]
+   [:ShutdownService request-shutdown shutdown-on-error]]
   (init [this context]
 
         (doseq [{:keys [reporter]} (vals metrics/metrics-registries)]
           (jmx-reporter/start reporter))
-        (assoc context :url-prefix (atom nil)))
+        (assoc context
+               :url-prefix (atom nil)
+               ;; This coordination is needed until we no longer
+               ;; support jdk < 10.
+               ;; https://bugs.openjdk.java.net/browse/JDK-8176254
+               :stop-status (atom #{})))
   (start [this context]
          (try+
-          (start-puppetdb context (get-config) this get-registered-endpoints)
+          (let [config (get-config)
+                upgrade? (get-in config [:global :upgrade-and-exit?])
+                context (start-puppetdb context config this
+                                        get-registered-endpoints upgrade?)]
+            (when upgrade?
+              ;; start-puppetdb has finished the migrations, which is
+              ;; all we needed to do.
+              (request-shutdown))
+            context)
           (catch [:type ::unsupported-database] {:keys [current oldest]}
-            (let [msg (db-unsupported-msg current oldest)
-                  attn (utils/attention-msg msg)]
-              (utils/println-err attn)
-              (log/error attn)
+            (let [msg (db-unsupported-msg current oldest)]
+              (utils/println-err (str (trs "error: ") msg))
+              (log/error msg)
               ;; Until TK-445 is resolved, we won't be able to avoid the
               ;; backtrace on exit this causes.
               (shutdown-on-error (service-id this)
@@ -484,12 +555,27 @@
   (clean [this] (clean this #{}))
   (clean [this what] (clean-puppetdb (service-context this) (get-config) what)))
 
-(def ^{:arglists `([& args])
-       :doc "Starts PuppetDB as a service via Trapperkeeper.  Aguments
-        TK's normal config parsing to do a bit of extra
-        customization."}
-  -main
-  (fn [& args]
-    (rh/add-hook #'puppetlabs.trapperkeeper.config/parse-config-data
-                 #'conf/hook-tk-parse-config-data)
-    (apply tk/main args)))
+(defn provide-services
+  "Starts PuppetDB as a service via Trapperkeeper.  Augments TK's normal
+  config parsing a bit."
+  ([args] (provide-services args nil))
+  ([args {:keys [upgrade-and-exit?]}]
+   (let [hook (if upgrade-and-exit?
+                (fn [f args]
+                  (assoc-in (#'conf/hook-tk-parse-config-data f args)
+                            [:global :upgrade-and-exit?]
+                            true))
+                #'conf/hook-tk-parse-config-data)]
+     (rh/add-hook #'puppetlabs.trapperkeeper.config/parse-config-data hook))
+   (apply tk/main args)
+   0))
+
+(defn cli
+  "Runs the services command as directed by the command line args and
+  returns an appropriate exit status."
+  ([args] (cli args nil))
+  ([args {:keys [upgrade-and-exit?] :as opts}]
+   (run-tk-cli-cmd #(provide-services args opts))))
+
+(defn -main [& args]
+  (exit (provide-services args)))
