@@ -10,6 +10,9 @@
    the migration function will be invoked, and the schema version and current
    time will be recorded in the schema_migrations table.
 
+   A migration function can return a map with ::vacuum-analyze to indicate what tables
+   need to be analyzed post-migration.
+
    NOTE: in order to support bug-fix schema changes to older branches without
    breaking the ability to upgrade, it is possible to define a sequence of
    migrations with non-sequential integers.  e.g., if the 1.0.x branch
@@ -56,10 +59,14 @@
             [puppetlabs.puppetdb.scf.storage-utils :as sutils]
             [clojure.set :refer :all]
             [puppetlabs.puppetdb.time :refer [to-timestamp]]
-            [clj-time.core :refer [now]]
+            [clj-time.core :refer [now interval in-millis]]
             [puppetlabs.puppetdb.jdbc :as jdbc :refer [query-to-vec]]
             [puppetlabs.puppetdb.config :as conf]
-            [puppetlabs.i18n.core :refer [trs]]))
+            [puppetlabs.i18n.core :refer [trs]]
+            [puppetlabs.puppetdb.scf.hash :as hash]
+            [clojure.set :as set]
+            [clojure.string :as str])
+  (:import [org.postgresql.util PGobject]))
 
 (defn init-through-2-3-8
   []
@@ -1281,6 +1288,7 @@
      (str "ALTER TABLE ONLY edges"
           "  ADD CONSTRAINT edges_certname_source_target_type_unique_key"
           "  UNIQUE (certname, source, target, type)"))))
+
 (defn add-latest-report-timestamp-to-certnames []
   (jdbc/do-commands
     "DROP INDEX IF EXISTS idx_certnames_latest_report_timestamp"
@@ -1288,6 +1296,206 @@
     "ALTER TABLE certnames ADD COLUMN latest_report_timestamp timestamp with time zone"
     "CREATE INDEX idx_certnames_latest_report_timestamp ON certnames(latest_report_timestamp)"))
 
+(defn reports-partial-indices
+  []
+  (jdbc/do-commands
+   "create index reports_noop_idx on reports(noop) where noop = true"
+   "create index reports_cached_catalog_status_on_fail on reports(cached_catalog_status) where cached_catalog_status = 'on_failure'"))
+
+(defn add-job-id []
+  (jdbc/do-commands
+   "alter table reports add column job_id text default null"
+   "alter table catalogs add column job_id text default null"
+   "create index reports_job_id_idx on reports(job_id) where job_id is not null"
+   "create index catalogs_job_id_idx on catalogs(job_id) where job_id is not null"))
+
+(defn maybe-parse-json [s]
+  (some-> s json/parse))
+
+(defn rededuplicate-facts []
+  (log/info (trs "[1/8] Cleaning up unreferenced facts..."))
+  (jdbc/do-commands
+   "DELETE FROM facts WHERE factset_id NOT IN (SELECT id FROM factsets)")
+
+  (log/info (trs "[2/8] Creating new fact storage tables..."))
+  (jdbc/do-commands
+   "CREATE SEQUENCE fact_values_id_seq;"
+
+   ;; will add not-null constraint on value_hash below; we don't have all the
+   ;; values for this yet
+   "CREATE TABLE fact_values (
+      id bigint DEFAULT nextval('fact_values_id_seq'::regclass) NOT NULL,
+      value_hash bytea,
+      value_type_id bigint NOT NULL,
+      value_integer bigint,
+      value_float double precision,
+      value_string text,
+      value_boolean boolean,
+      value jsonb
+    );"
+
+   ;; Do this early to help with the value_hash update queries below
+   "ALTER TABLE fact_values ADD CONSTRAINT fact_values_pkey PRIMARY KEY (id);"
+
+   "CREATE TABLE facts_transform (
+       factset_id bigint NOT NULL,
+       fact_path_id bigint NOT NULL,
+       fact_value_id bigint NOT NULL
+    );")
+
+  (log/info (trs "[3/8] Copying unique fact values into fact_values"))
+  (jdbc/do-commands
+   "INSERT INTO fact_values (value, value_integer, value_float, value_string, value_boolean, value_type_id)
+       SELECT distinct value, value_integer, value_float, value_string, value_boolean, value_type_id FROM facts")
+
+
+  ;; Handle null fv.value separately; allowing them here leads to an intractable
+  ;; query plan
+  (log/info (trs "[4/8] Reconstructing facts to refer to fact_values..."))
+  (jdbc/do-commands
+   "INSERT INTO facts_transform (factset_id, fact_path_id, fact_value_id)
+       SELECT f.factset_id, f.fact_path_id, fv.id
+         FROM facts f
+        INNER JOIN fact_values fv
+                ON fv.value_type_id = f.value_type_id
+               AND fv.value = f.value
+               AND fv.value_integer IS NOT DISTINCT FROM f.value_integer
+               AND fv.value_float   IS NOT DISTINCT FROM f.value_float
+               AND fv.value_string  IS NOT DISTINCT FROM f.value_string
+               AND fv.value_boolean IS NOT DISTINCT FROM f.value_boolean
+        WHERE f.value IS NOT NULL AND fv.value IS NOT NULL"
+
+   "INSERT INTO facts_transform (factset_id, fact_path_id, fact_value_id)
+       SELECT f.factset_id, f.fact_path_id, fv.id
+         FROM facts f
+        INNER JOIN fact_values fv
+                ON fv.value_type_id = f.value_type_id
+               AND fv.value IS NOT DISTINCT FROM f.value
+        WHERE f.value IS NULL AND fv.value IS NULL")
+
+  (log/info (trs "[5/8] Cleaning up duplicate null values..."))
+  ;; only do this if the DB has some null values
+  (when (pos? (-> (jdbc/query-to-vec "select count(*) from fact_values
+                                      where value_type_id in (4,5)
+                                      and (value = 'null' or value is null)")
+                  first
+                  :count))
+    (let [existing-id (some-> (jdbc/query-to-vec
+                               "select id from fact_values
+                                where value_type_id=4 and value is null")
+                              first
+                              :id)
+          real-null-value-id (or existing-id
+                                 (-> (jdbc/query-to-vec
+                                      "insert into fact_values (value_type_id, value) values (4, null) returning id")
+                                     first
+                                     :id))
+          bad-null-value-ids (->> (jdbc/query-to-vec "select id from fact_values
+                                                      where (value_type_id = 5 and value = 'null')
+                                                         or (value_type_id = 5 and value is null)
+                                                         or (value_type_id = 4 and value = 'null')")
+                                 (map :id))]
+     (doseq [id bad-null-value-ids]
+       (jdbc/do-prepared "update facts_transform set fact_value_id = (?) where fact_value_id = (?)"
+                         [real-null-value-id id])
+       (jdbc/do-prepared "delete from fact_values where id = (?)"
+                         [id]))))
+
+  (log/info (trs "[6/8] Computing fact value hashes..."))
+  (jdbc/call-with-query-rows
+   ["select id, value::text from fact_values"]
+   (fn [rows]
+     (doseq [batch (partition-all 500 rows)]
+       (let [ids (map :id batch)
+             hashes (map #(-> (:value %)
+                              maybe-parse-json
+                              hash/generic-identity-hash
+                              sutils/munge-hash-for-storage)
+                         batch)]
+         (jdbc/do-prepared
+          "update fact_values set value_hash = in_data.hash
+            from (select unnest(?) as id, unnest(?) as hash) in_data
+            where fact_values.id = in_data.id"
+          [(sutils/array-to-param "bigint" Long ids)
+           (sutils/array-to-param "bytea" PGobject hashes)])))))
+
+  (log/info (trs "[7/8] Indexing fact_values table..."))
+  (jdbc/do-commands
+   "DROP TABLE facts"
+   "ALTER TABLE facts_transform rename to facts"
+
+   "ALTER TABLE fact_values alter column value_hash set not null"
+   "ALTER TABLE fact_values ADD CONSTRAINT fact_values_value_hash_key UNIQUE (value_hash);"
+   "CREATE INDEX fact_values_value_float_idx ON fact_values USING btree (value_float);"
+   "CREATE INDEX fact_values_value_integer_idx ON fact_values USING btree (value_integer);")
+
+  (log/info (trs "[8/8] Indexing facts table..."))
+  (jdbc/do-commands
+   "ALTER TABLE facts ADD CONSTRAINT facts_factset_id_fact_path_id_fact_key UNIQUE (factset_id, fact_path_id);"
+   "CREATE INDEX facts_fact_path_id_idx ON facts USING btree (fact_path_id);"
+   "CREATE INDEX facts_fact_value_id_idx ON facts USING btree (fact_value_id);"
+
+   "ALTER TABLE facts ADD CONSTRAINT fact_path_id_fk
+      FOREIGN KEY (fact_path_id)
+      REFERENCES fact_paths(id);"
+
+   "ALTER TABLE facts ADD CONSTRAINT fact_value_id_fk
+      FOREIGN KEY (fact_value_id)
+      REFERENCES fact_values(id)
+      ON UPDATE RESTRICT
+      ON DELETE RESTRICT;"
+
+   "ALTER TABLE facts ADD CONSTRAINT factset_id_fk FOREIGN KEY (factset_id) REFERENCES factsets(id) ON UPDATE CASCADE ON DELETE CASCADE;")
+
+  {::vacuum-analyze #{"facts" "fact_values" "fact_paths"}})
+
+(defn varchar-columns-to-text []
+  (jdbc/do-commands
+    "alter table reports
+     alter column puppet_version type text,
+     alter column configuration_version type text"
+
+    "alter table resource_events
+     alter column status type text,
+     alter column property type text,
+     alter column containing_class type text,
+     alter column file type text"))
+
+(defn jsonb-facts []
+  (jdbc/do-commands
+   "alter table factsets add column paths_hash bytea"  ;; Can be null atm
+   "alter table factsets add column stable jsonb"
+   "alter table factsets add column stable_hash bytea"
+   "alter table factsets add column volatile jsonb"
+   "create index idx_factsets_jsonb_merged on factsets using gin((stable||volatile) jsonb_path_ops)"
+
+   "update factsets fs
+    set stable = (select json_object_agg(name, value)
+                 from (
+                 select f.factset_id, fp.name, fv.value from facts f
+                 inner join fact_values fv on fv.id = f.fact_value_id
+                 inner join fact_paths fp on fp.id = f.fact_path_id
+                 inner join value_types vt on vt.id = fv.value_type_id
+                 where fp.depth = 0
+                 ) s where fs.id = s.factset_id),
+    volatile = jsonb('{}')"
+
+   "drop table facts"
+   "drop table fact_values"
+   "alter table fact_paths drop constraint fact_paths_path_key"
+
+   ;; TODO consider migrating fact paths - maybe not worth it. This table will
+   ;; be mostly repopulated on reception of first factset, and fully
+   ;; repopulated by the time runinterval has elasped. It also only matters to
+   ;; the fact-paths endpoint now.
+   "truncate table fact_paths"
+
+   "alter table fact_paths add column path_array text[] not null"
+   "alter table fact_paths add column value_type_id int not null"
+   "alter table fact_paths add constraint fact_paths_path_type_unique unique(path, value_type_id)")
+
+  {::vacuum-analyze #{"factsets"}})
+
 (def migrations
   "The available migrations, as a map from migration version to migration function."
   {28 init-through-2-3-8
@@ -1326,7 +1534,12 @@
    58 add-gin-index-on-resource-params-cache
    59 improve-facts-factset-id-index
    60 fix-missing-edges-fk-constraint
-   61 add-latest-report-timestamp-to-certnames})
+   61 add-latest-report-timestamp-to-certnames
+   62 reports-partial-indices
+   63 add-job-id
+   64 rededuplicate-facts
+   65 varchar-columns-to-text
+   66 jsonb-facts})
 
 (def desired-schema-version (apply max (keys migrations)))
 
@@ -1351,13 +1564,10 @@
           results (jdbc/with-db-transaction []  (query-to-vec query))]
       (apply sorted-set (map :version results)))
     (catch java.sql.SQLException e
-      (let [message (.getMessage e)
-            sql-state (.getSQLState e)]
-        (if (and (or (= sql-state "42P01") ; postgresql: undefined_table
-                     (= sql-state "42501")) ; hsqldb: user lacks privilege or object not found
-                 (re-find #"(?i)schema_migrations" message))
-          (sorted-set)
-          (throw e))))))
+      (let [sql-state (.getSQLState e)]
+        (if (= sql-state "42P01") ; postgresql: undefined_table
+            (sorted-set)
+            (throw e))))))
 
 (defn pending-migrations
   "Returns a collection of pending migrations, ordered from oldest to latest."
@@ -1432,12 +1642,21 @@
                    unexpected))))
 
     (if-let [pending (seq (pending-migrations))]
-      (do
-        (jdbc/with-db-transaction []
-          (doseq [[version migration] pending]
-            (log/info (trs "Applying database migration version {0}" version))
-            (sql-or-die (fn [] (migration) (record-migration! version)))))
-        (sutils/analyze-small-tables small-tables)
+      (let [tables-to-analyze (jdbc/with-db-transaction []
+                                (->> pending
+                                     (map (fn [[version migration]]
+                                            (let [t0 (now)]
+                                            (log/info (trs "Applying database migration version {0}" version))
+                                            (sql-or-die (fn [] (let [result (migration)]
+                                                                    (record-migration! version)
+                                                                    result)))
+                                            (log/info (trs "Applied database migration version {0} in {1} ms"
+                                                           version (in-millis (interval t0 (now))))))))
+                                     (filter map?)
+                                     (map ::vacuum-analyze)
+                                     (apply set/union small-tables)
+                                     sort))]
+        (log/info (trs "Updating table statistics for: {0}" (str/join ", " tables-to-analyze)))
         true)
       (do
         (log/info (trs "There are no pending migrations"))
@@ -1452,13 +1671,6 @@
     (log/info (trs "Creating additional index `fact_paths_path_trgm`"))
     (jdbc/do-commands
      "CREATE INDEX fact_paths_path_trgm ON fact_paths USING gist (path gist_trgm_ops)"))
-  (jdbc/do-commands
-   "drop index if exists fact_values_string_trgm")
-  (when-not (sutils/index-exists? "facts_value_string_trgm")
-    (log/info (trs "Creating additional index `facts_value_string_trgm`"))
-    (jdbc/do-commands
-     ["create index facts_value_string_trgm on facts"
-      "  using gin (value_string gin_trgm_ops)"]))
 
   (when-not (sutils/index-exists? "packages_name_trgm")
     (log/info (trs "Creating additional index `packages_name_trgm`"))
@@ -1466,6 +1678,18 @@
      ["create index packages_name_trgm on packages"
       "  using gin (name gin_trgm_ops)"])))
 
+(defn ensure-report-id-index []
+  (when-not (sutils/index-exists? "idx_reports_compound_id")
+    (log/info "Indexing reports for id queries")
+
+    (jdbc/do-commands
+     "create index idx_reports_compound_id on reports
+        (producer_timestamp, certname, hash)
+        where start_time is not null")
+
+    (jdbc/do-commands-outside-txn
+     "vacuum analyze reports")))
+
 (defn indexes!
   "Create missing indexes for applicable database platforms."
   [config]
@@ -1478,4 +1702,5 @@
         "\n\n"
         (trs "We are unable to create the recommended pg_trgm indexes due to\nthe extension not being installed correctly.")
         " "
-        (trs " Run the command:\n\n    CREATE EXTENSION pg_trgm;\n\nas the database super user on the PuppetDB database to correct\nthis, then restart PuppetDB.\n"))))))
+        (trs " Run the command:\n\n    CREATE EXTENSION pg_trgm;\n\nas the database super user on the PuppetDB database to correct\nthis, then restart PuppetDB.\n")))))
+  (ensure-report-id-index))
