@@ -22,7 +22,7 @@
             [puppetlabs.puppetdb.reports :as reports]
             [puppetlabs.puppetdb.facts :as facts :refer [facts-schema]]
             [puppetlabs.kitchensink.core :as kitchensink]
-            [puppetlabs.kitchensink.core :as ks]
+            [puppetlabs.puppetdb.scf.storage-utils :as sutils]
             [com.rpl.specter :as sp]
             [clojure.java.jdbc :as sql]
             [clojure.set :as set]
@@ -43,13 +43,14 @@
             [puppetlabs.puppetdb.jdbc :as jdbc :refer [query-to-vec]]
             [puppetlabs.puppetdb.time :refer [to-timestamp]]
             [honeysql.core :as hcore]
-            [puppetlabs.i18n.core :refer [trs]])
-  (:import [org.postgresql.util PGobject]
+            [puppetlabs.i18n.core :refer [trs]]
+            [puppetlabs.puppetdb.package-util :as pkg-util]
+            [puppetlabs.puppetdb.cheshire :as json])
+  (:import [java.security MessageDigest]
+           [java.util Arrays]
+           [org.postgresql.util PGobject]
            [org.joda.time Period]))
 
-(defn sql-ts-after? [^java.sql.Timestamp x ^java.sql.Timestamp y]
-  (.after x y))
-
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Schemas
 
@@ -239,13 +240,30 @@
    ["SELECT expired FROM certnames WHERE certname=?" certname]
    (comp :expired first sql/result-set-seq)))
 
-(defn purge-deactivated-and-expired-nodes!
-  "Delete nodes from the database which were deactivated before `time`."
-  [time]
-  {:pre [(kitchensink/datetime? time)]}
-  (let [ts (to-timestamp time)]
-    (jdbc/delete! :certname_packages ["certname_id in (select id from certnames where deactivated < ? OR expired < ?)" ts ts])
-    (jdbc/delete! :certnames ["deactivated < ? OR expired < ?" ts ts])))
+(defn-validated purge-deactivated-and-expired-nodes!
+  "Delete nodes from the database which were deactivated before horizon."
+  ([horizon :- (s/pred kitchensink/datetime?)]
+   (let [ts (to-timestamp horizon)]
+     (jdbc/do-prepared
+      (str "with ids as (delete from certnames"
+           "               where deactivated < ? or expired < ?"
+           "               returning id)"
+           "  delete from certname_packages"
+           "    where certname_id in (select * from ids)")
+      [ts ts])))
+  ([horizon batch-limit]
+   {:pre [(kitchensink/datetime? horizon)]}
+   (let [ts (to-timestamp horizon)]
+     (jdbc/do-prepared
+      (str "with ids as (delete from certnames"
+           "               where id in (select id from certnames"
+           "                              where deactivated < ?"
+           "                                    or expired < ?"
+           "                                limit ?)"
+           "               returning id)"
+           "  delete from certname_packages"
+           "    where certname_id in (select * from ids)")
+      [ts ts batch-limit]))))
 
 (defn activate-node!
   "Reactivate the given host. Adds the host to the database if it was not
@@ -374,6 +392,7 @@
            resources
            version
            code_id
+           job_id
            transaction_uuid
            catalog_uuid
            environment
@@ -386,6 +405,7 @@
    :catalog_uuid (sutils/munge-uuid-for-storage catalog_uuid)
    :timestamp (to-timestamp received-timestamp)
    :code_id code_id
+   :job_id job_id
    :environment_id (ensure-environment environment)
    :producer_timestamp (to-timestamp producer_timestamp)
    :producer_id (ensure-producer producer)
@@ -419,7 +439,7 @@
   (if (seq resource-hashes)
     (let [resource-array (->> (vec resource-hashes)
                               (map sutils/munge-hash-for-storage)
-                              (sutils/array-to-param "bytea" PGobject))
+                              (sutils/array-to-param "bytea" org.postgresql.util.PGobject))
           query (format "SELECT DISTINCT %s as resource FROM resource_params_cache WHERE resource=ANY(?)"
                         (sutils/sql-hash-as-str "resource"))
           sql-params [query resource-array]]
@@ -765,7 +785,7 @@
                    latest-producer-timestamp :producer_timestamp} (latest-catalog-metadata certname)]
               (cond
                 (some-> latest-producer-timestamp
-                        (sql-ts-after? (to-timestamp producer_timestamp)))
+                        (.after (to-timestamp producer_timestamp)))
                 (log/warn (trs "Not replacing catalog for certname {0} because local data is newer." certname))
 
                 (= stored-hash hash)
@@ -819,186 +839,10 @@
    (:gc-packages performance-metrics)
    (jdbc/delete!
     :packages
-    ["id NOT IN (SELECT package_id from certname_packages)"])))
+    ["not exists (select * from certname_packages cp where cp.package_id = packages.id)"])))
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
-;;; Facts
 
-(defn-validated certname-to-factset-id :- s/Int
-  "Given a certname, returns the factset id."
-  [certname :- String]
-  (jdbc/query-with-resultset
-   ["SELECT id from factsets WHERE certname = ?" certname]
-   (comp :id first sql/result-set-seq)))
-
-(defn-validated certname-factset-metadata :- {:package_hash (s/maybe s/Str)
-                                              :factset_id s/Int
-                                              :certname_id s/Int}
-  "Given a certname, return the factset id, hash and certname id."
-  [certname :- s/Str]
-  (jdbc/query-with-resultset
-   [(format "SELECT fs.id as factset_id, c.id as certname_id, %s as package_hash
-             FROM factsets fs, certnames c
-             WHERE fs.certname = ? AND c.certname = ?"
-            (sutils/sql-hash-as-str "c.package_hash"))
-    certname certname]
-   (comp first sql/result-set-seq)))
-
-(defn-validated delete-pending-path-id-orphans!
-  "Delete paths in dropped-pids that are no longer mentioned
-   in other factsets."
-  [factset-id dropped-pids]
-  (when-let [dropped-pids (seq dropped-pids)]
-    (let [dropped-chunks (partition-all gc-chunksize dropped-pids)
-          candidate-chunks (map (partial sutils/array-to-param "bigint" Long)
-                                dropped-chunks)]
-      (dorun
-       (map (fn [candidate-paths]
-              (jdbc/do-prepared
-               "DELETE FROM fact_paths fp
-                        WHERE fp.id = ANY (?)
-                        AND NOT EXISTS (SELECT 1 FROM facts f
-                                        WHERE f.fact_path_id = ANY(?)
-                                        AND f.fact_path_id = fp.id
-                                        AND f.factset_id <> ?)"
-               (concat [candidate-paths]
-                       [candidate-paths]
-                       [factset-id])))
-            candidate-chunks)))))
-
-(defn-validated delete-orphaned-paths! :- s/Int
-  "Deletes up to n paths that are no longer mentioned by any factsets,
-  and returns the number actually deleted.  These orphans can be
-  created by races between parallel updates since (for performance) we
-  don't serialize those transactions.  Via repeatable read, an update
-  transaction may decide not to delete paths that are only referred to
-  by other facts that are being changed in parallel transactions to
-  also not refer to the paths."
-  [n :- (s/constrained s/Int (complement neg?))]
-  (if (zero? n)
-    0
-    (first
-     (jdbc/with-db-transaction []
-       (jdbc/do-prepared
-        "DELETE FROM fact_paths
-           WHERE id IN (SELECT fp.id FROM fact_paths fp
-                          WHERE NOT EXISTS (SELECT 1
-                                              FROM facts f
-                                              WHERE fp.id = f.fact_path_id)
-                          LIMIT ?)"
-        [n])))))
-
-;; NOTE: now only used in tests.
-(defn-validated delete-certname-facts!
-  "Delete all the facts for certname."
-  [certname :- String]
-  (jdbc/with-db-transaction []
-    (let [factset-id (certname-to-factset-id certname)
-          dead-pids
-          (jdbc/call-with-query-rows
-           ["select fact_path_id from facts where factset_id = ?"
-            factset-id]
-           {:as-arrays? true}
-           #(mapv first (rest %)))]
-      (jdbc/do-commands
-       (format "delete from facts where factset_id = %s" factset-id))
-      (delete-pending-path-id-orphans! factset-id dead-pids)
-      (jdbc/delete! :factsets ["id=?" factset-id]))))
-
-(defn realize-paths!
-  "Ensures that all paths exist in the database and returns a map of
-  paths to ids."
-  [pathstrs]
-  (when-let [pathstrs (seq pathstrs)]
-    (let [array-to-param sutils/array-to-param
-          existing-path-ids (jdbc/call-with-query-rows
-                             [(str "select path, id from fact_paths"
-                                   "  where path in (select * from unnest(?))")
-                              (array-to-param "text" String pathstrs)]
-                             {:as-arrays? true}
-                             (fn [[col-names & rows]]
-                               (into {} rows)))
-          missing-db-paths (set/difference (set pathstrs)
-                                           (set (keys existing-path-ids)))
-          new-paths (map facts/string-to-factpath missing-db-paths)]
-
-      (jdbc/call-with-query-rows
-       [(str "insert into fact_paths (path, name, depth)"
-             "  (select * from unnest(?, ?, ?)) returning path, id")
-        (array-to-param "text" String (map facts/factpath-to-string new-paths))
-        (array-to-param "text" String (map first new-paths))
-        (array-to-param "integer" Integer (map #(-> (count %) dec int)
-                                               new-paths))]
-       {:as-arrays? true}
-       (fn [[col-names & rows]]
-         (into existing-path-ids rows))))))
-
-(def ^:const db-fv-string 0)
-(def ^:const db-fv-int 1)
-(def ^:const db-fv-float 2)
-(def ^:const db-fv-bool 3)
-(def ^:const db-fv-nil 4)
-(def ^:const db-fv-struct 5)
-
-(defn fact-value-type-id [x]
-  (cond
-    (keyword? x) db-fv-string
-    (string? x) db-fv-string
-    (integer? x) db-fv-int
-    (float? x) db-fv-float
-    (kitchensink/boolean? x) db-fv-bool
-    (nil? x) db-fv-nil
-    (coll? x) db-fv-struct
-    :else (throw (Exception. (trs "Fact value {0} has unexpected type"
-                                  (pr-str x))))))
-
-(def value-row-cols
-  ;; Must match order in value->value-row immediately below
-  [:value_type_id
-   :large_value_hash
-   :value
-   :value_string
-   :value_integer
-   :value_float
-   :value_boolean])
-
-(defn storage-hash [x]
-  ;; If the return type changes here, update the PGobject check in update-facts.
-  (-> x shash/generic-identity-hash sutils/munge-hash-for-storage))
-
-(defn value->value-row
-  ([value] (value->value-row value nil))
-  ([value hash]
-   (let [jsonb (when-not (nil? value)
-                 (sutils/munge-jsonb-for-storage value))
-         large? (and jsonb (>= (count (.getValue ^PGobject jsonb))
-                               facts/large-value-threshold))
-         hash (when large?
-                (or hash (storage-hash value)))]
-     (value->value-row value hash jsonb)))
-  ([value hash jsonb]
-   (let [id (fact-value-type-id value)]
-     (condp = id
-       db-fv-string [id hash jsonb value nil nil nil]
-       db-fv-int    [id hash jsonb nil value nil nil]
-       db-fv-float  [id hash jsonb nil nil value nil]
-       db-fv-bool   [id hash jsonb nil nil nil value]
-       db-fv-struct [id hash jsonb nil nil nil nil]
-       db-fv-nil    [id hash jsonb nil nil nil nil]
-       (throw (Exception. (trs "Unexpected type {0} for {1}"
-                               (pr-str id) (pr-str value))))))))
-
-(defn insert-facts! [factset-id pids values]
-  (when (seq values)
-    (let [cols (concat [:factset_id :fact_path_id] value-row-cols)]
-      (jdbc/insert-multi! :facts
-                          cols
-                          (map (fn [pid value]
-                                 (apply vector
-                                        factset-id pid
-                                        (value->value-row value)))
-                               pids
-                               values)))))
+;;; Packages
 
 (defn find-certname-id-and-hash [certname]
   (first (jdbc/query-to-vec [(format "SELECT id, %s as package_hash FROM certnames WHERE certname=?"
@@ -1032,7 +876,7 @@
 
 (defn insert-missing-packages [existing-hashes-map new-hashed-package-tuples]
   (let [packages-to-create (remove (fn [hashed-package-tuple]
-                                     (get existing-hashes-map (facts/package-tuple-hash hashed-package-tuple)))
+                                     (get existing-hashes-map (pkg-util/package-tuple-hash hashed-package-tuple)))
                                    new-hashed-package-tuples)
         results (jdbc/insert-multi! :packages
                                     (map (fn [[package_name version provider package-hash]]
@@ -1042,7 +886,7 @@
                                             :hash (sutils/munge-hash-for-storage package-hash)})
                                          packages-to-create))]
     (merge existing-hashes-map
-           (zipmap (map facts/package-tuple-hash packages-to-create)
+           (zipmap (map pkg-util/package-tuple-hash packages-to-create)
                    (map :id results)))))
 
 (s/defn update-packages
@@ -1050,12 +894,12 @@
   `certname`. Differences will result in updates to the database"
   [certname_id :- s/Int
    package_hash :- (s/maybe s/Str)
-   inventory :- [facts/package-tuple]]
+   inventory :- [pkg-util/package-tuple]]
   (let [hashed-package-tuples (map shash/package-identity-hash inventory)
         new-package-hash (shash/package-similarity-hash hashed-package-tuples)]
 
     (when-not (= new-package-hash package_hash)
-      (let [just-hashes (map facts/package-tuple-hash hashed-package-tuples)
+      (let [just-hashes (map pkg-util/package-tuple-hash hashed-package-tuples)
             existing-package-hashes (find-package-hashes just-hashes)
             full-hashes-map (insert-missing-packages existing-package-hashes hashed-package-tuples)
             new-package-id-set (set (vals full-hashes-map))
@@ -1090,7 +934,7 @@
   (let [certname-id (:id (find-certname-id-and-hash certname))
         hashed-package-tuples (map shash/package-identity-hash inventory)
         new-packageset-hash (shash/package-similarity-hash hashed-package-tuples)
-        just-hashes (map facts/package-tuple-hash hashed-package-tuples)
+        just-hashes (map pkg-util/package-tuple-hash hashed-package-tuples)
         existing-package-hashes (find-package-hashes just-hashes)
         ;; a map of package hash to id in the packages table
         full-hashes-map (insert-missing-packages existing-package-hashes hashed-package-tuples)
@@ -1105,6 +949,76 @@
                         (map (fn [package-id] [certname-id package-id])
                              (vals full-hashes-map)))))
 
+
+;;; JSONB facts
+
+(defn-validated certname-factset-metadata
+  :- {:package_hash (s/maybe s/Str)
+      :factset_id s/Int
+      :certname_id s/Int
+      :stable_hash (s/maybe (s/pred shash/sha1-bytes?))}
+  "Given a certname, return the factset id, hash and certname id."
+  [certname :- s/Str]
+  (jdbc/query-with-resultset
+   [(format "SELECT fs.id as factset_id, c.id as certname_id, %s as package_hash, fs.stable_hash as stable_hash
+             FROM factsets fs, certnames c
+             WHERE fs.certname = ? AND c.certname = ?"
+            (sutils/sql-hash-as-str "c.package_hash"))
+    certname certname]
+   (comp first sql/result-set-seq)))
+
+
+(defn volatile-fact-keys-for-factset [factset-id]
+  (->> (jdbc/query-to-vec "select jsonb_object_keys(volatile) as fact from factsets where id=?"
+                          factset-id)
+       (map :fact)))
+
+(defn load-stable-facts [factset-id]
+  (-> (jdbc/query-to-vec "select stable from factsets where id=?"
+                         factset-id)
+      first
+      :stable
+      str
+      json/parse-string))
+
+;; Chunk size for path insertion to avoid blowing prepared statement size limit
+(def path-insertion-chunk-size 6000)
+
+(defn pathmap-digestor [digest]
+  (fn [{:keys [path value_type_id] :as pathmap}]
+    (.update digest (-> (str path value_type_id)
+                        (.getBytes "UTF-8")))
+    pathmap))
+
+(defn realize-paths
+  "Ensures that every path in the pathmaps has a corresponding row in
+  fact_paths, and returns either nil, if pathmaps is empty, or a
+  fingerprint of the paths if hash? is true."
+  ([pathmaps] (realize-paths pathmaps identity))
+  ([pathmaps notice-pathmap]
+   (let [path-array-conversion #(->> (map str %)
+                                     (sutils/array-to-param "text" String))]
+     (when (seq pathmaps)
+       (->> pathmaps
+            (map notice-pathmap)
+            (map #(update % :path_array path-array-conversion))
+            (partition-all path-insertion-chunk-size)
+            (map #(jdbc/insert-multi! :fact_paths % {:on-conflict "do nothing"}))
+            dorun)))))
+
+(defn hash-pathmaps-paths [pathmaps]
+  (let [digest (MessageDigest/getInstance "SHA-1")]
+    (->> pathmaps
+         (map (pathmap-digestor digest))
+         dorun)
+    (.digest digest)))
+
+(defn factset-paths-hash [factset_id]
+  (-> "select paths_hash from factsets where id = ?"
+      (query-to-vec factset_id)
+      first
+      :paths_hash))
+
 (pls/defn-validated add-facts!
   "Given a certname and a map of fact names to values, store records for those
   facts associated with the certname."
@@ -1113,170 +1027,133 @@
      :as fact-data} :- facts-schema
     include-hash? :- s/Bool]
    (jdbc/with-db-transaction []
-     (jdbc/insert!
-      :factsets
-      (merge
-       {:certname certname
-        :timestamp (to-timestamp timestamp)
-        :environment_id (ensure-environment environment)
-        :producer_timestamp (to-timestamp producer_timestamp)
-        :producer_id (ensure-producer producer)}
-       (when include-hash?
-         {:hash (sutils/munge-hash-for-storage
-                 (shash/fact-identity-hash fact-data))})))
-    (let [paths-and-values (facts/facts->paths-and-values values)
-          pathstrs (map (comp facts/factpath-to-string first) paths-and-values)
-          paths->ids (realize-paths! pathstrs)
-          path-values (map second paths-and-values)]
-
-      (when (seq package_inventory)
-        (insert-packages certname package_inventory))
-
-      (insert-facts! (certname-to-factset-id certname)
-                     (map paths->ids pathstrs)
-                     path-values)))))
-
-(defn- update-existing-values [factset-id change-info]
-  (when (seq change-info)
-    (sql/db-do-prepared
-     jdbc/*db* true
-     (cons
-      (str "update facts set "
-           (str/join ", " (map (fn [col] (str (name col) " = ?"))
-                              value-row-cols))
-           "  where factset_id = ? and fact_path_id = ?")
-      (map (fn [[_ path path-id new-val new-hash]]
-             (conj (if new-hash
-                     (value->value-row new-val new-hash)
-                     (value->value-row new-val))
-                   factset-id
-                   path-id))
-           change-info))
-     {:multi? true})))
-
-(defn compare-fv-to-row
-  "Returns a map with a logically false :changed? value if new-val
-  does not differ from the row data, and a logically true value
-  otherwise.  If a hash of new-val was computed during the
-  process and changed? is true, it will be returned in :hash."
-  [new-val type-id hash v vstr vint vfloat vbool]
-  (if (not= type-id (fact-value-type-id new-val))
-    {:changed? true}
-    (cond
-      ;; faster direct compare
-      (or (= type-id db-fv-int)
-          (= type-id db-fv-float)
-          (= type-id db-fv-bool))
-      {:changed? (not= new-val (or vint vfloat vbool))}
-
-      ;; nil is a singleton
-      (= type-id db-fv-nil) {:changed? false}
-
-      ;; things that may have a hash
-      (or (= type-id db-fv-string)
-          (= type-id db-fv-struct))
-      (if hash
-        (let [new-hash (storage-hash new-val)]
-          (if (= hash new-hash)
-            {:changed? false}
-            {:changed? true :hash new-hash}))
-        (cond
-          (= type-id db-fv-string)
-          {:changed? (not= new-val vstr)}
-          (= type-id db-fv-struct)
-          {:changed? (not= new-val (sutils/parse-db-json v))}
-          :else
-          (throw (Exception. (trs "Unexpected type {0} for {1}"
-                                  (pr-str type-id) (pr-str new-val))))))
-      :else
-      (throw (Exception. (trs "Unexpected type {0} for {1}"
-                              (pr-str type-id) (pr-str new-val)))))))
+     (let [paths-hash (let [digest (MessageDigest/getInstance "SHA-1")]
+                        (realize-paths (facts/facts->pathmaps values)
+                                       (pathmap-digestor digest))
+                        (.digest digest))]
+       (jdbc/insert! :factsets
+                     (merge
+                      {:certname certname
+                       :timestamp (to-timestamp timestamp)
+                       :environment_id (ensure-environment environment)
+                       :producer_timestamp (to-timestamp producer_timestamp)
+                       :producer_id (ensure-producer producer)
+                       :stable (sutils/munge-jsonb-for-storage values)
+                       :stable_hash (shash/generic-identity-sha1-bytes values)
+                       ;; need at least an empty map for the jsonb || operator
+                       :volatile (sutils/munge-jsonb-for-storage {})
+                       :paths_hash paths-hash}
+                      (when include-hash?
+                        {:hash (sutils/munge-hash-for-storage
+                                (shash/fact-identity-hash fact-data))}))))
+     (when (seq package_inventory)
+       (insert-packages certname package_inventory)))))
 
 (s/defn update-facts!
   "Given a certname, querys the DB for existing facts for that
    certname and will update, delete or insert the facts as necessary
    to match the facts argument. (cf. add-facts!)"
-  [{:keys [certname values environment timestamp producer_timestamp producer package_inventory]
-    :as fact-data} :- facts-schema]
+  [{:keys [certname values environment timestamp producer_timestamp producer package_inventory] :as fact-data}
+   :- facts-schema]
   (jdbc/with-db-transaction []
-    (let [{:keys [package_hash certname_id factset_id]} (certname-factset-metadata certname)
-          paths-and-values (facts/facts->paths-and-values values)
-          pathstrs (map (comp facts/factpath-to-string first) paths-and-values)
-          pathstr->value (into {} (map (fn [pathstr [_ v]] [pathstr v])
-                                       pathstrs
-                                       paths-and-values))
-          existing-status ;; :changed :unchanged :gone
-          (jdbc/call-with-query-rows
-           [(str "select"
-                 "    path, id, value_type_id, large_value_hash,"
-                 "    case when large_value_hash is null "
-                 "              and value_type_id = " db-fv-struct
-                 "           then value"
-                 "    end as value,"
-                 "    case when large_value_hash is null then value_string"
-                 "    end as value_string,"
-                 "    value_integer, value_float, value_boolean"
-                 "  from facts as f"
-                 "  inner join fact_paths fp on f.fact_path_id = fp.id"
-                 "  where f.factset_id = ?")
-            factset_id]
-           {:as-arrays? true}
-           (fn [[col-names & rows]]
-             (doall
-              (map (fn [[path id type-id hash v vstr vint vfloat vbool]]
-                     ;; Classify row as :changed :unchanged: :gone
-                     (let [new-val (pathstr->value path ::not-found)]
-                       (if (= new-val ::not-found)
-                         [:gone path id]
-                         (let [{:keys [changed? hash]}
-                               (compare-fv-to-row new-val
-                                                  type-id hash
-                                                  v vstr vint vfloat vbool)]
-                           (if-not changed?
-                             [:unchanged path id]
-                             (if hash
-                               [:changed path id new-val hash]
-                               [:changed path id new-val]))))))
-                   rows))))
-          rm-pids (seq (for [[st _ pid] existing-status
-                             :when (= :gone st)]
-                         pid))
-          new-pathstrs (let [have? (set (for [[st path] existing-status
-                                              :when #(#{:changed :unchanged} st)]
-                                          path))]
-                         (remove have? pathstrs))
-          new-values (map pathstr->value new-pathstrs)
-          paths->ids (realize-paths! new-pathstrs)]
+    (let [{:keys [package_hash certname_id factset_id stable_hash volatile_fact_names]}
+          (certname-factset-metadata certname)
 
-      ;; Paths are unique per factset so we can delete solely based on pid.
-      (when rm-pids
-        (jdbc/do-prepared
-         (format "delete from facts where factset_id = ? and fact_path_id %s"
-                 (jdbc/in-clause rm-pids))
-         (cons factset_id rm-pids)))
+          ;; split facts into stable and volatile maps. Everything that was
+          ;; previously volatile stays that way. Any stable value that changes
+          ;; becomes volatile. Newly volatile facts are removed from the stable
+          ;; map.
+          current-volatile-fact-keys (volatile-fact-keys-for-factset factset_id)
+          incoming-volatile-facts (select-keys values current-volatile-fact-keys)
+          incoming-stable-facts (apply dissoc values current-volatile-fact-keys)
+          incoming-stable-fact-hash (shash/generic-identity-sha1-bytes incoming-stable-facts)
+          fact-json-updates (if-not (Arrays/equals stable_hash incoming-stable-fact-hash)
+                              ;; stable facts are different; load the json and move the ones that
+                              ;; changed into volatile
+                              (let [current-stable-facts (load-stable-facts factset_id)
+                                    changed-facts (->> incoming-stable-facts
+                                                       (filter (fn [[fact new-value]]
+                                                                 (when-let [current (get current-stable-facts fact)]
+                                                                   (not= new-value current))))
+                                                       (into {}))
+                                    new-stable-facts (->> incoming-stable-facts
+                                                          (remove (fn [[fact _]] (get changed-facts fact)))
+                                                          (into {}))
+                                    new-volatile-facts (merge incoming-volatile-facts
+                                                              changed-facts)]
+                                {:stable (sutils/munge-jsonb-for-storage new-stable-facts)
+                                 :stable_hash (shash/generic-identity-sha1-bytes new-stable-facts)
+                                 :volatile (sutils/munge-jsonb-for-storage new-volatile-facts)})
 
-      (insert-facts! factset_id
-                     (map paths->ids new-pathstrs)
-                     new-values)
-
-      (update-existing-values factset_id
-                              (filter #(= :changed (first %))
-                                      existing-status))
-
-      (delete-pending-path-id-orphans! factset_id rm-pids)
+                              ;; no change to stable facts, so just update volatile
+                              {:volatile (sutils/munge-jsonb-for-storage incoming-volatile-facts)})]
 
       (when (or package_hash (seq package_inventory))
         (update-packages certname_id package_hash package_inventory))
 
-      (jdbc/update! :factsets
-                    {:timestamp (to-timestamp timestamp)
-                     :environment_id (ensure-environment environment)
-                     :producer_timestamp (to-timestamp producer_timestamp)
-                     :hash (-> fact-data
-                               shash/fact-identity-hash
-                               sutils/munge-hash-for-storage)
-                     :producer_id (ensure-producer producer)}
-                    ["id=?" factset_id]))))
+      ;; Only update the paths if any existing paths_hash doesn't
+      ;; match the incoming paths.
+      (let [paths-hash (if-let [existing-hash (factset-paths-hash factset_id)]
+                         (let [incoming-hash (-> (facts/facts->pathmaps values)
+                                                 hash-pathmaps-paths)]
+                           (if (Arrays/equals existing-hash incoming-hash)
+                             existing-hash
+                             (do
+                               (realize-paths (facts/facts->pathmaps values))
+                               incoming-hash)))
+                         ;; No existing hash
+                         (let [digest (MessageDigest/getInstance "SHA-1")]
+                           (realize-paths (facts/facts->pathmaps values)
+                                          (pathmap-digestor digest))
+                           (.digest digest)))]
+
+        (jdbc/update! :factsets
+                      (merge
+                       {:timestamp (to-timestamp timestamp)
+                        :environment_id (ensure-environment environment)
+                        :producer_timestamp (to-timestamp producer_timestamp)
+                        :hash (-> fact-data
+                                  shash/fact-identity-hash
+                                  sutils/munge-hash-for-storage)
+                        :producer_id (ensure-producer producer)
+                        :paths_hash paths-hash}
+                       fact-json-updates)
+                      ["id=?" factset_id])))))
+
+(defn delete-unused-fact-paths []
+  "Deletes paths from fact_paths that are no longer needed by any
+  factset.  In the unusual case where a path changes type, the
+  previous version will linger.  This requires a parent transaction
+  with at least a repeatable-read isolation level (and may or may not
+  need postgres' \"stronger than the standard\" repeatable-read
+  behavior).  Otherwise paths could be added to fact_paths elsewhere
+  during the gc, not be noticed, and then be deleted at the end."
+  ;; Use a temp table for now until we figure out why pg is creating a
+  ;; vast number of temp files when this is all handled as a single
+  ;; query.  (PDB-3924)
+  (jdbc/do-commands
+   ["with recursive live_paths(path, value) as"
+    "   (select key as path, value"
+    "      from (select (jsonb_each(stable||volatile)).* from factsets) as base_case"
+    "      union"
+    "        select path||'#~'||sub_level.key as path,"
+    "               sub_level.value"
+    "          from live_paths,"
+    "          lateral (select *"
+    "                     from (select (jsonb_each(value)).*"
+    "                             where jsonb_typeof(value) = 'object') as sub_fields"
+    "                     union (select generate_series(0, jsonb_array_length(value - 1))::text as key,"
+    "                                   jsonb_array_elements(value) as value"
+    "                              where jsonb_typeof(value) = 'array')) as sub_level)"
+    "   select path into unlogged tmp_live_paths from live_paths"]
+
+   "analyze tmp_live_paths"
+
+   ["delete from fact_paths fp"
+    "  where not exists (select 1 from tmp_live_paths"
+    "                      where tmp_live_paths.path = fp.path)"]
+
+   "drop table tmp_live_paths"))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Reports
@@ -1337,13 +1214,21 @@
     (dissoc row-map :environment_id)
     row-map))
 
+(defn replace-null-bytes [x]
+  (if-not (string? x)
+    x
+    (let [^String s x]
+      (if (= -1 (.indexOf s (int \u0000)))
+        s
+        (.replace s \u0000 \ufffd)))))
+
 (defn normalize-resource-event
   "Prep `event` for comparison/computation of a hash"
   [event]
   (-> event
       (update :timestamp to-timestamp)
-      (update :old_value sutils/db-serialize)
-      (update :new_value sutils/db-serialize)
+      (update :old_value (comp sutils/db-serialize replace-null-bytes))
+      (update :new_value (comp sutils/db-serialize replace-null-bytes))
       (assoc :containing_class (find-containing-class (:containment_path event)))))
 
 (defn normalize-report
@@ -1369,7 +1254,7 @@
          (let [{:keys [puppet_version certname report_format configuration_version producer
                        producer_timestamp start_time end_time transaction_uuid environment
                        status noop metrics logs resources resource_events catalog_uuid
-                       code_id cached_catalog_status noop_pending corrective_change]
+                       code_id job_id cached_catalog_status noop_pending corrective_change]
                 :as report} (normalize-report orig-report)
                 report-hash (shash/report-identity-hash report)]
            (jdbc/with-db-transaction []
@@ -1382,6 +1267,7 @@
                                 :transaction_uuid (sutils/munge-uuid-for-storage transaction_uuid)
                                 :catalog_uuid (sutils/munge-uuid-for-storage catalog_uuid)
                                 :code_id code_id
+                                :job_id job_id
                                 :cached_catalog_status cached_catalog_status
                                 :metrics (sutils/munge-jsonb-for-storage metrics)
                                 :logs (sutils/munge-jsonb-for-storage logs)
@@ -1429,7 +1315,7 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Public
 
-(def oldest-supported-db [9 4])
+(def oldest-supported-db [9 6])
 
 (pls/defn-validated add-certname!
   "Add the given host to the db"
@@ -1452,7 +1338,7 @@
     (boolean
      (some-> entity
              (timestamp-of-newest-record certname)
-             (sql-ts-after? time)))))
+             (.after time)))))
 
 (pls/defn-validated have-newer-record-for-certname?
   [certname :- String
@@ -1528,8 +1414,7 @@
   [{:keys [certname producer_timestamp] :as fact-data} :- facts-schema]
   (time! (:replace-facts performance-metrics)
          (if-let [local-factset-producer-ts (timestamp-of-newest-record :factsets certname)]
-           (if-not (sql-ts-after? local-factset-producer-ts
-                                  (to-timestamp producer_timestamp))
+           (if-not (.after local-factset-producer-ts (to-timestamp producer_timestamp))
              (update-facts! fact-data)
              (log/warn (trs "Not updating facts for certname {0} because local data is newer." certname)))
            (add-facts! fact-data))))
@@ -1540,17 +1425,16 @@
    received-timestamp :- pls/Timestamp]
   (add-report!* report received-timestamp true))
 
-(def ^:dynamic *orphaned-path-gc-limit* 200)
-
 (defn garbage-collect!
   "Delete any lingering, unassociated data in the database"
   [db]
   (time!
    (:gc performance-metrics)
-   (jdbc/with-transacted-connection db
-     (delete-unassociated-params!)
-     (delete-unassociated-environments!))
-   ;; These require serializable because they make the decision to
-   ;; delete based on row counts in another table.
-   (jdbc/with-transacted-connection' db :serializable
-     (delete-orphaned-paths! *orphaned-path-gc-limit*))))
+   (do
+     (jdbc/with-transacted-connection db
+       (delete-unassociated-params!)
+       (delete-unassociated-environments!))
+     (jdbc/with-transacted-connection' db :repeatable-read
+       ;; May or may not require postgresql's "stronger than the
+       ;; standard" behavior for repeatable read.
+       (delete-unused-fact-paths)))))
