@@ -1,6 +1,7 @@
 (ns puppetlabs.puppetdb.command-test
   (:require [me.raynes.fs :as fs]
             [clojure.java.jdbc :as sql]
+            [clojure.string :as str]
             [metrics.counters :as counters]
             [metrics.meters :as meters]
             [puppetlabs.puppetdb.cheshire :as json]
@@ -20,10 +21,11 @@
             [puppetlabs.puppetdb.scf.hash :as shash]
             [puppetlabs.puppetdb.testutils.db :refer [*db* with-test-db]]
             [schema.core :as s]
-            [puppetlabs.trapperkeeper.testutils.logging :refer [atom-logger]]
+            [puppetlabs.trapperkeeper.testutils.logging
+             :refer [atom-logger logs-matching with-log-output]]
             [clj-time.format :as tfmt]
             [puppetlabs.puppetdb.cli.services :as cli-svc]
-            [puppetlabs.puppetdb.command :refer :all]
+            [puppetlabs.puppetdb.command :as cmd :refer :all]
             [puppetlabs.puppetdb.config :as conf]
             [puppetlabs.puppetdb.reports :as reports]
             [puppetlabs.puppetdb.testutils
@@ -34,7 +36,7 @@
             [puppetlabs.puppetdb.jdbc-test :refer [full-sql-exception-msg]]
             [puppetlabs.puppetdb.examples :refer :all]
             [puppetlabs.puppetdb.testutils.services :as svc-utils
-             :refer [*server*]]
+             :refer [*server* with-pdb-with-no-gc]]
             [puppetlabs.puppetdb.command.constants :refer [command-names]]
             [clj-time.coerce
              :refer [from-sql-date to-timestamp to-date-time to-string]]
@@ -45,7 +47,7 @@
             [slingshot.test]
             [puppetlabs.puppetdb.utils :as utils]
             [puppetlabs.puppetdb.time :as pt]
-            [puppetlabs.trapperkeeper.app :refer [get-service]]
+            [puppetlabs.trapperkeeper.app :refer [get-service app-context]]
             [clojure.core.async :as async]
             [puppetlabs.kitchensink.core :as ks]
             [clojure.string :as str]
@@ -58,7 +60,9 @@
             [puppetlabs.trapperkeeper.services
              :refer [service-context]]
             [overtone.at-at :refer [mk-pool scheduled-jobs]]
-            [puppetlabs.puppetdb.testutils :as tu])
+            [puppetlabs.puppetdb.testutils :as tu]
+            [puppetlabs.puppetdb.client :as client]
+            [puppetlabs.puppetdb.threadpool :as gtp])
   (:import [java.nio.file Files]
            [java.util.concurrent TimeUnit]
            [org.joda.time DateTime DateTimeZone]))
@@ -71,6 +75,12 @@
     (fs/delete-dir (:path dlo))
     (#'overtone.at-at/shutdown-pool-now! @(:pool-atom delay-pool))))
 
+;; define blacklist map to allow use of with-redefs later
+(def blacklist-config {:facts-blacklist [#"blacklisted-fact"
+                                         #"pre.*" #".*suff"
+                                         #"gl.?b" #"p[u].*et"]
+                          :facts-blacklist-type "regex"})
+
 (defn create-message-handler-context [q]
   (let [delay-pool (mk-pool)
         command-chan (async/chan 10)
@@ -82,7 +92,16 @@
                              (:registry (new-metrics "puppetlabs.puppetdb.dlo"
                                                      :jmx? false)))]
     (map->CommandHandlerContext
-     {:handle-message (message-handler q command-chan dlo delay-pool *db* response-chan stats)
+     {:handle-message (message-handler
+                       q
+                       command-chan
+                       dlo
+                       delay-pool
+                       *db*
+                       response-chan
+                       stats
+                       blacklist-config
+                       (atom {:executing-delayed 0}))
       :command-chan command-chan
       :dlo dlo
       :delay-pool delay-pool
@@ -132,6 +151,14 @@
   (queue/create-command-req "replace catalog" version certname nil "" identity
                             (tqueue/coerce-to-stream payload)))
 
+(s/defn function-that-needs-int-via-schema [x :- s/Int]
+  x)
+
+(defn function-that-needs-int-via-precondition [x]
+  {:pre [(integer? x)]}
+  x)
+
+
 (deftest command-processor-integration
   (let [v5-catalog (get-in wire-catalogs [5 :empty])]
     (testing "correctly formed messages"
@@ -153,6 +180,26 @@
               (is (= 0 (count (scheduled-jobs delay-pool))))
               (is (= 2 (count (fs/list-dir (:path dlo))))))))
 
+        (testing "when a schema error occurs should be discarded to the dead letter queue"
+          (with-redefs [process-command-and-respond! (fn [& _] (upon-error-throw-fatality
+                                                                (function-that-needs-int-via-schema "string")))]
+            (with-message-handler {:keys [handle-message dlo delay-pool q]}
+              (let [discards (discard-count)]
+                (handle-message (queue/store-command q (catalog->command-req 5 v5-catalog)))
+                (is (= (inc discards) (discard-count))))
+              (is (= 0 (count (scheduled-jobs delay-pool))))
+              (is (= 2 (count (fs/list-dir (:path dlo))))))))
+
+        (testing "when a precondition error occurs should be discarded to the dead letter queue"
+          (with-redefs [process-command-and-respond! (fn [& _] (upon-error-throw-fatality
+                                                                (function-that-needs-int-via-precondition "string")))]
+            (with-message-handler {:keys [handle-message dlo delay-pool q]}
+              (let [discards (discard-count)]
+                (handle-message (queue/store-command q (catalog->command-req 5 v5-catalog)))
+                (is (= (inc discards) (discard-count))))
+              (is (= 0 (count (scheduled-jobs delay-pool))))
+              (is (= 2 (count (fs/list-dir (:path dlo))))))))
+
         (testing "when a non-fatal error occurs should be requeued with the error recorded"
           (let [expected-exception (Exception. "non-fatal error")]
             (with-redefs [process-command-and-respond! (fn [& _]
@@ -361,14 +408,18 @@
             (is (= 0 (task-count delay-pool)))
             (is (empty? (fs/list-dir (:path dlo))))))
 
-        (testing "with code-id should store the catalog"
-          (with-message-handler {:keys [handle-message dlo delay-pool q]}
-            (handle-message
-             (queue/store-command q (catalog->command-req version-num (assoc catalog :code_id "my_git_sha1"))))
-            (is (= [(with-env {:certname certname :code_id "my_git_sha1"})]
-                   (query-to-vec "SELECT certname, code_id, environment_id FROM catalogs")))
-            (is (= 0 (task-count delay-pool)))
-            (is (empty? (fs/list-dir (:path dlo))))))
+        (testing "with optional ids should store the catalog"
+          (let [job-id (when (> version-num 8) "1337")]
+            (with-message-handler {:keys [handle-message dlo delay-pool q]}
+              (handle-message
+                (queue/store-command q (catalog->command-req
+                                         version-num
+                                         (cond-> (assoc catalog :code_id "my_git_sha1")
+                                           job-id (assoc :job_id job-id)))))
+              (is (= [(with-env {:certname certname :code_id "my_git_sha1" :job_id job-id})]
+                     (query-to-vec "SELECT certname, code_id, job_id, environment_id FROM catalogs")))
+              (is (= 0 (task-count delay-pool)))
+              (is (empty? (fs/list-dir (:path dlo)))))))
 
         (testing "with an existing catalog should replace the catalog"
           (with-message-handler {:keys [handle-message dlo delay-pool q]}
@@ -642,6 +693,35 @@
   "Support fact command versions"
   [:v4])
 
+(defn query-factsets [& cols]
+  (->> (str "select certname,"
+            "       (stable||volatile) as facts,"
+            "       environment,"
+            "       producer_timestamp"
+            "  from factsets
+                 inner join environments on factsets.environment_id=environments.id")
+       query-to-vec
+       (map #(update % :facts (comp json/parse-string str)))
+       (map #(select-keys % cols))))
+
+(defn query-facts
+        [& cols]
+        (->> (str "select certname,"
+                  "       environment,"
+                  "       value #>> '{}' as value,"
+                  "       key as name,"
+                  "       producer_timestamp"
+                  "  from (select certname,"
+                  "               producer_timestamp,"
+                  "               environment_id,"
+                  "               (jsonb_each((stable||volatile))).*"
+                  "          from factsets) fs"
+                  ;; Q: is this wrong?  fs.key outside subselect?
+                  "    left join environments env on fs.environment_id = env.id"
+                  "    order by fs.key")
+             query-to-vec
+             (map #(select-keys % cols))))
+
 (let [certname  "foo.example.com"
       facts     {:certname certname
                  :environment "DEV"
@@ -649,35 +729,83 @@
                           "b" "2"
                           "c" "3"}
                  :producer_timestamp (to-timestamp (now))}
-      v4-command {:command (command-names :replace-facts)
-                  :version 4
-                  :payload facts}
       one-day   (* 24 60 60 1000)
       yesterday (to-timestamp (- (System/currentTimeMillis) one-day))
       tomorrow  (to-timestamp (+ (System/currentTimeMillis) one-day))]
 
+  (deftest facts-package-deduplication
+    (let [version :v5
+          command (assoc facts
+                         :producer "node0"
+                         :package_inventory [["foo" "1.2.3" "apt"]
+                                             ["bar" "2.3.4" "apt"]
+                                             ["bar" "2.3.4" "apt"]
+                                             ["baz" "3.4.5" "apt"]])]
+      (testing "should deduplicate package_inventory on facts v5"
+        (with-message-handler {:keys [handle-message dlo delay-pool q]}
+          (handle-message (queue/store-command q (facts->command-req (version-kwd->num version) command)))
+          (is (= [["bar" "2.3.4" "apt"]
+                  ["baz" "3.4.5" "apt"]
+                  ["foo" "1.2.3" "apt"]]
+                 (rest
+                  (jdbc/query
+                   ["SELECT p.name as package_name, p.version, p.provider
+                     FROM certname_packages cp
+                          inner join packages p on cp.package_id = p.id
+                          inner join certnames c on cp.certname_id = c.id
+                     WHERE c.certname = ?
+                     ORDER BY package_name, version, provider"
+                    certname]
+                   {:as-arrays? true}))))))))
+
+  (deftest regex-facts-blacklist
+    (dotestseq [version fact-versions
+                :let [command (update facts :values
+                                      #(assoc %
+                                              "blacklisted-fact" "val"
+                                              "prefix" "val"
+                                              "facts-suff" "val"
+                                              "glob" "val"
+                                              "puppet" "val"
+                                              ;; facts below shouldn't be blacklisted
+                                              "notprefix" "val"
+                                              "suff-not" "val"
+                                              "not-blacklisted-fact" "val"))]]
+      (testing "should ignore the blacklisted facts using regex"
+        (with-message-handler {:keys [handle-message dlo delay-pool q]}
+          (handle-message (queue/store-command q (facts->command-req (version-kwd->num version) command)))
+          (is (= [{:certname certname :name "a" :value "1"}
+                  {:certname certname :name "b" :value "2"}
+                  {:certname certname :name "c" :value "3"}
+                  {:certname certname :name "not-blacklisted-fact" :value "val"}
+                  {:certname certname :name "notprefix" :value "val"}
+                  {:certname certname :name "suff-not" :value "val"}]
+                 (query-facts :certname :name :value)))))))
+
+    (deftest literal-facts-blacklist
+        (dotestseq [version fact-versions
+                    :let [command (update facts :values
+                                          #(assoc % "blacklisted-fact" "val"))]]
+          (testing "should ignore the blacklisted fact"
+            (with-redefs [blacklist-config
+                          {:facts-blacklist ["blacklisted-fact"]
+                           :facts-blacklist-type "literal"}]
+              (with-message-handler {:keys [handle-message dlo delay-pool q]}
+                (handle-message (queue/store-command q (facts->command-req (version-kwd->num version) command)))
+                (is (= [{:certname certname :name "a" :value "1"}
+                        {:certname certname :name "b" :value "2"}
+                        {:certname certname :name "c" :value "3"}]
+                       (query-facts :certname :name :value))))))))
+
   (deftest replace-facts-no-facts
     (dotestseq [version fact-versions
                 :let [command facts]]
       (testing "should store the facts"
         (with-message-handler {:keys [handle-message dlo delay-pool q]}
           (handle-message (queue/store-command q (facts->command-req (version-kwd->num version) command)))
-          (is (= (query-to-vec
-                  "SELECT fp.path as name,
-                          COALESCE(f.value_string,
-                                   cast(f.value_integer as text),
-                                   cast(f.value_boolean as text),
-                                   cast(f.value_float as text),
-                                   '') as value,
-                          fs.certname
-                   FROM factsets fs
-                     INNER JOIN facts as f on fs.id = f.factset_id
-                     INNER JOIN fact_paths as fp on f.fact_path_id = fp.id
-                   WHERE fp.depth = 0
-                   ORDER BY name ASC")
-                 [{:certname certname :name "a" :value "1"}
-                  {:certname certname :name "b" :value "2"}
-                  {:certname certname :name "c" :value "3"}]))
+          (is (= [{:certname certname
+                   :facts {"a" "1", "b" "2", "c" "3"}}]
+                 (query-factsets :certname :facts)))
           (is (= 0 (task-count delay-pool)))
           (is (empty? (fs/list-dir (:path dlo))))
           (let [result (query-to-vec "SELECT certname,environment_id FROM factsets")]
@@ -707,22 +835,9 @@
                         yesterday))
               (is (= (scf-store/environment-id "DEV") (:environment_id result))))
 
-            (is (= (query-to-vec
-                    "SELECT fp.path as name,
-                          COALESCE(f.value_string,
-                                   cast(f.value_integer as text),
-                                   cast(f.value_boolean as text),
-                                   cast(f.value_float as text),
-                                   '') as value,
-                          fs.certname
-                   FROM factsets fs
-                     INNER JOIN facts as f on fs.id = f.factset_id
-                     INNER JOIN fact_paths as fp on f.fact_path_id = fp.id
-                   WHERE fp.depth = 0
-                   ORDER BY fp.path ASC")
-                   [{:certname certname :name "a" :value "1"}
-                    {:certname certname :name "b" :value "2"}
-                    {:certname certname :name "c" :value "3"}]))
+            (is (= [{:certname certname
+                     :facts {"a" "1", "b" "2", "c" "3"}}]
+                   (query-factsets :certname :facts)))
             (is (= 0 (task-count delay-pool)))
             (is (empty? (fs/list-dir (:path dlo)))))))))
 
@@ -744,19 +859,10 @@
 
           (is (= (query-to-vec "SELECT certname,timestamp,environment_id FROM factsets")
                  [(with-env {:certname certname :timestamp tomorrow})]))
-          (is (= (query-to-vec
-                  "SELECT fp.path as name,
-                          COALESCE(f.value_string,
-                                   cast(f.value_integer as text),
-                                   cast(f.value_boolean as text),
-                                   cast(f.value_float as text),
-                                   '') as value,
-                          fs.certname
-                   FROM factsets fs
-                     INNER JOIN facts as f on fs.id = f.factset_id
-                     INNER JOIN fact_paths as fp on f.fact_path_id = fp.id
-                   WHERE fp.depth = 0
-                   ORDER BY name ASC")
+          (is (= [{:certname certname
+                   :facts {"x" "24", "y" "25", "z" "26"}}]
+                 (query-factsets :certname :facts)))
+          (is (= (query-facts :certname :name :value)
                  [{:certname certname :name "x" :value "24"}
                   {:certname certname :name "y" :value "25"}
                   {:certname certname :name "z" :value "26"}]))
@@ -774,19 +880,10 @@
           (handle-message (queue/store-command q (facts->command-req (version-kwd->num version) command)))
           (is (= (query-to-vec "SELECT certname,deactivated FROM certnames")
                  [{:certname certname :deactivated nil}]))
-          (is (= (query-to-vec
-                  "SELECT fp.path as name,
-                          COALESCE(f.value_string,
-                                   cast(f.value_integer as text),
-                                   cast(f.value_boolean as text),
-                                   cast(f.value_float as text),
-                                   '') as value,
-                          fs.certname
-                   FROM factsets fs
-                     INNER JOIN facts as f on fs.id = f.factset_id
-                     INNER JOIN fact_paths as fp on f.fact_path_id = fp.id
-                   WHERE fp.depth = 0
-                   ORDER BY name ASC")
+          (is (= [{:certname certname
+                   :facts {"a" "1", "b" "2", "c" "3"}}]
+                 (query-factsets :certname :facts)))
+          (is (= (query-facts :certname :name :value)
                  [{:certname certname :name "a" :value "1"}
                   {:certname certname :name "b" :value "2"}
                   {:certname certname :name "c" :value "3"}]))
@@ -803,19 +900,10 @@
 
           (is (= (query-to-vec "SELECT certname,deactivated FROM certnames")
                  [{:certname certname :deactivated tomorrow}]))
-          (is (= (query-to-vec
-                  "SELECT fp.path as name,
-                          COALESCE(f.value_string,
-                                   cast(f.value_integer as text),
-                                   cast(f.value_boolean as text),
-                                   cast(f.value_float as text),
-                                   '') as value,
-                          fs.certname
-                   FROM factsets fs
-                     INNER JOIN facts as f on fs.id = f.factset_id
-                     INNER JOIN fact_paths as fp on f.fact_path_id = fp.id
-                   WHERE fp.depth = 0
-                   ORDER BY name ASC")
+          (is (= [{:certname certname
+                   :facts {"a" "1", "b" "2", "c" "3"}}]
+                 (query-factsets :certname :facts)))
+          (is (= (query-facts :certname :name :value)
                  [{:certname certname :name "a" :value "1"}
                   {:certname certname :name "b" :value "2"}
                   {:certname certname :name "c" :value "3"}]))
@@ -840,22 +928,23 @@
     (with-message-handler {:keys [handle-message dlo delay-pool q]}
       (handle-message (queue/store-command q (facts->command-req 3 facts-cmd)))
 
-      (is (= (query-to-vec
-              "SELECT fp.path as name,
-                          COALESCE(f.value_string,
-                                   cast(f.value_integer as text),
-                                   cast(f.value_boolean as text),
-                                   cast(f.value_float as text),
-                                   '') as value,
-                          fs.certname,
-                          e.environment,
-                          fs.producer_timestamp
-                   FROM factsets fs
-                     INNER JOIN facts as f on fs.id = f.factset_id
-                     INNER JOIN fact_paths as fp on f.fact_path_id = fp.id
-                     INNER JOIN environments as e on fs.environment_id = e.id
-                   WHERE fp.depth = 0
-                   ORDER BY name ASC")
+      (is (= [{:certname certname
+               :facts {"a" "1", "b" "2", "c" "3"}
+               :producer_timestamp producer-time
+               :environment "DEV"}]
+             (query-factsets :certname :facts :producer_timestamp :environment)))
+      #_(is (= (query-to-vec (str "select certname,"
+                                  "       environment,"
+                                  "       value,"
+                                  "       key as name,"
+                                  "       producer_timestamp"
+                                  "  from (select certname,"
+                                  "               producer_timestamp,"
+                                  "               environment_id,"
+                                  "               (jsonb_each((stable||volatile))).*"
+                                  "          from factsets) fs"
+                                  "            left join environments env on fs.environment_id = env.id"
+                                  "          order by fs.key"))
              [{:certname certname :name "a" :value "1" :producer_timestamp producer-time :environment "DEV"}
               {:certname certname :name "b" :value "2" :producer_timestamp producer-time :environment "DEV"}
               {:certname certname :name "c" :value "3" :producer_timestamp producer-time :environment "DEV"}]))
@@ -876,21 +965,10 @@
 
       (handle-message (queue/store-command q (facts->command-req 2 facts-cmd)))
 
-      (is (= (query-to-vec
-              "SELECT fp.path as name,
-                          COALESCE(f.value_string,
-                                   cast(f.value_integer as text),
-                                   cast(f.value_boolean as text),
-                                   cast(f.value_float as text),
-                                   '') as value,
-                          fs.certname,
-                          e.environment
-                   FROM factsets fs
-                     INNER JOIN facts as f on fs.id = f.factset_id
-                     INNER JOIN fact_paths as fp on f.fact_path_id = fp.id
-                     INNER JOIN environments as e on fs.environment_id = e.id
-                   WHERE fp.depth = 0
-                   ORDER BY name ASC")
+      (is (= [{:certname certname
+               :facts {"a" "1", "b" "2", "c" "3"}}]
+             (query-factsets :certname :facts)))
+      (is (= (query-facts :certname :name :value :environment)
              [{:certname certname :name "a" :value "1" :environment "DEV"}
               {:certname certname :name "b" :value "2" :environment "DEV"}
               {:certname certname :name "c" :value "3" :environment "DEV"}]))
@@ -919,7 +997,7 @@
                                                                            ""
                                                                            identity
                                                                            (tqueue/coerce-to-stream "bad stuff"))))
-          (is (empty? (query-to-vec "SELECT * FROM facts")))
+          (is (empty? (query-to-vec "SELECT * FROM factsets")))
           (is (= 0 (task-count delay-pool)))
           (is (seq (fs/list-dir (:path dlo)))))))))
 
@@ -933,7 +1011,7 @@
                                                                        ""
                                                                        identity
                                                                        (tqueue/coerce-to-stream "bad stuff"))))
-      (is (empty? (query-to-vec "SELECT * FROM facts")))
+      (is (empty? (query-to-vec "SELECT * FROM factsets")))
       (is (= 0 (task-count delay-pool)))
       (is (seq (fs/list-dir (:path dlo)))))))
 
@@ -1286,6 +1364,48 @@
 (defn- get-config []
   (conf/get-config (get-service svc-utils/*server* :DefaultedConfig)))
 
+(deftest bashed-commands-handled-correctly
+  (let [real-dochan gtp/dochan
+        go-ahead-and-execute (promise)]
+    (with-redefs [gtp/dochan (fn [& args]
+                               @go-ahead-and-execute
+                               (apply real-dochan args))]
+      (svc-utils/with-puppetdb-instance
+        (let [{pdb-host :host pdb-port :port
+               :or {pdb-host "127.0.0.1" pdb-port 8080}} (:jetty (get-config))
+              base-url (utils/pdb-cmd-base-url pdb-host pdb-port :v1)
+              facts {:certname "foo.com"
+                     :environment "test"
+                     :producer_timestamp (str (now))
+                     :producer "puppetserver"
+                     :values {:foo "1"
+                              :bar "2"}
+                     :package_inventory [["openssl" "1.1.0e-1" "apt"]]}
+              dispatcher (get-service svc-utils/*server* :PuppetDBCommandDispatcher)
+              response-mult (response-mult dispatcher)
+              response-watch-ch (async/chan 10)
+              _ (async/tap response-mult response-watch-ch)]
+
+          ;; submit two replace facts commands for the same certname
+          ;; this should cause one of the commands to get "bashed" in the queue
+          (client/submit-facts base-url "foo.com" 5 facts)
+          ;; make sure the second command has a later timestamp
+          (Thread/sleep 200)
+          (client/submit-facts base-url "foo.com" 5 facts)
+          ;; allow pdb to process messages
+          (deliver go-ahead-and-execute true)
+
+          ;; grab reponse maps from commands above, timeout if not received
+          (let [[val _] (async/alts!! [(async/into
+                                        #{} (async/take 2 response-watch-ch))
+                                       (async/timeout tu/default-timeout-ms)])]
+            (when-not val
+              (throw (Exception. "timed out waiting for response-chan")))
+
+            ;; check the first command was "bashed"
+            (is (= #{{:id 0, :delete? true} {:id 1, :delete? nil}}
+                   (set (map #(select-keys % [:id :delete?]) val))))))))))
+
 (deftest command-service-stats
   (svc-utils/with-puppetdb-instance
     (let [pdb (get-service svc-utils/*server* :PuppetDBServer)
@@ -1532,3 +1652,53 @@
             ;; fast enough, initially (at least), the rates can be zero.
             (is (= (+ 3 (:total start-global)) (:total global)))
             (is (= (+ 3 (or (:total start-cmd) 0)) (:total cmd)))))))))
+
+
+;; Test mitigation of
+;; https://bugs.openjdk.java.net/browse/JDK-8176254 i.e. handling of
+;; an unexpected in-flight garbage collection (via the scheduling
+;; bug) during stop.  For now, just test that if a gc is in flight,
+;; we wait on it if it doesn't take too long, and we proceed anyway
+;; if it does.
+
+(defn test-stop-with-delayed-commands [infinite-delay? expected-msg-rx]
+  (with-log-output log-output
+    (let [msg-count 3
+          enq-latch (java.util.concurrent.CountDownLatch. msg-count)
+          enq-proceed (promise)]
+      (with-redefs [cmd/schedule-msg-after (fn [ms f _] (f))  ;; no delay
+                    cmd/enqueue-delayed-message (fn [& args]
+                                                  (.countDown enq-latch)
+                                                  @enq-proceed
+                                                  ;; do nothing
+                                                  nil)]
+        (with-pdb-with-no-gc
+          (let [pdb (get-service *server* :PuppetDBServer)
+                q (-> pdb cli-svc/shared-globals :q)
+                dispatcher (get-service *server* :PuppetDBCommandDispatcher)
+                stop-status (-> dispatcher service-context :stop-status)
+                v5-catalog (get-in wire-catalogs [5 :empty])
+                cmdreqs (repeatedly msg-count #(catalog->command-req 5 v5-catalog))]
+            (doseq [cmdreq cmdreqs
+                    :let [cmdref (queue/store-command q cmdreq)
+                          {:keys [command version]} cmdref]]
+              ;; schedule-delayed-message assumes the metrics already exist
+              (cmd/create-metrics-for-command! command version)
+              (utils/noisy-future
+               (cmd/schedule-delayed-message cmdref (Exception.) :dummy-chan
+                                             :dummy-pool stop-status)))
+            (.await enq-latch)
+            (is #{:executing-delayed 3} @stop-status)
+            (when-not infinite-delay?
+              (deliver enq-proceed true))))))
+    (is (= 1 (->> @log-output
+                  (logs-matching expected-msg-rx)
+                  count)))))
+
+(deftest stop-waits-a-bit-for-delayed-cmds
+  (test-stop-with-delayed-commands false
+                                   #"^Halted delayed command processsing"))
+
+(deftest stop-ignores-delayed-cmds-that-take-too-long
+  (test-stop-with-delayed-commands true
+                                   #"^Forcibly terminating delayed command processing"))
