@@ -15,6 +15,7 @@
                                                            enumerated-resource-events-map]]
             [puppetlabs.puppetdb.testutils.http
              :refer [*app*
+                     are-error-response-headers
                      deftest-http-app
                      query-response
                      vector-param
@@ -26,8 +27,6 @@
 (def endpoints [[:v4 "/v4/events"]
                 [:v4 "/v4/environments/DEV/events"]])
 
-(def content-type-json http/json-response-content-type)
-
 (defn parse-result
   "Stringify (if needed) then parse the response"
   [body]
@@ -198,7 +197,7 @@
                                                                 [{:field "resource-title"}])})
               body (get response :body "null")]
           (is (= (:status response) http/status-bad-request))
-          (is (= (:headers response) {"Content-Type" http/error-response-content-type}))
+          (are-error-response-headers (:headers response))
           (is (re-find #"Unrecognized column 'resource-title' specified in :order_by" body)))))))
 
 (deftest-http-app query-distinct-resources
@@ -216,7 +215,7 @@
                                       {:distinct_resources true})
             body (get response :body "null")]
         (is (= (:status response) http/status-bad-request))
-        (is (= (:headers response) {"Content-Type" http/error-response-content-type}))
+        (are-error-response-headers (:headers response))
         (is (re-find
              #"'distinct_resources' query parameter requires accompanying parameters 'distinct_start_time' and 'distinct_end_time'"
              body)))
@@ -224,7 +223,7 @@
                                      {:distinct_resources true :distinct_start_time 0})
             body (get response :body "null")]
         (is (= (:status response) http/status-bad-request))
-        (is (= (:headers response) {"Content-Type" http/error-response-content-type}))
+        (are-error-response-headers (:headers response))
         (is (re-find
              #"'distinct_resources' query parameter requires accompanying parameters 'distinct_start_time' and 'distinct_end_time'"
              body)))
@@ -232,7 +231,7 @@
                                      {:distinct_resources true :distinct_end_time 0})
             body (get response :body "null")]
         (is (= (:status response) http/status-bad-request))
-        (is (= (:headers response) {"Content-Type" http/error-response-content-type}))
+        (are-error-response-headers (:headers response))
         (is (re-find
              #"'distinct_resources' query parameter requires accompanying parameters 'distinct_start_time' and 'distinct_end_time'"
              body)))
@@ -241,7 +240,7 @@
                                       {:distinct_start_time 0 :distinct_end_time 0})
             body      (get response :body "null")]
         (is (= (:status response) http/status-bad-request))
-        (is (= (:headers response) {"Content-Type" http/error-response-content-type}))
+        (are-error-response-headers (:headers response))
         (is (re-find
              #"'distinct_resources' query parameter must accompany parameters 'distinct_start_time' and 'distinct_end_time'"
              body))))
@@ -482,7 +481,7 @@
 
                  ["in" "certname" ["extract" ["nothing" "nothing2" "certname"] ["select_resources"
                                                                                 ["=" "type" "Class"]]]]
-                 #"Can't extract unknown 'resources' fields: 'nothing', 'nothing2'.*Acceptable fields are.*"
+                 #"Can't extract unknown 'resources' fields 'nothing' and 'nothing2'.*Acceptable fields are.*"
 
                  ;; In-query for invalid fields should throw an error
                  ["in" "nothing" ["extract" "certname" ["select_resources"
@@ -491,7 +490,7 @@
 
                  ["in" ["certname" "nothing" "nothing2"] ["extract" "certname" ["select_resources"
                                                                                 ["=" "type" "Class"]]]]
-                 #"Can't match on unknown 'events' fields: 'nothing', 'nothing2' for 'in'.*Acceptable fields are.*")))
+                 #"Can't match on unknown 'events' fields 'nothing' and 'nothing2' for 'in'.*Acceptable fields are.*")))
 
 (deftest-http-app invalid-subqueries
   [[version endpoint] endpoints
@@ -504,7 +503,7 @@
              method endpoint query)]
         (is (re-find msg body))
         (is (= status http/status-bad-request))
-        (is (= headers {"Content-Type" http/error-response-content-type}))))))
+        (are-error-response-headers headers)))))
 
 (def versioned-invalid-queries
   (omap/ordered-map
@@ -518,7 +517,7 @@
                    #"Can't extract unknown 'events' field 'nothing'.*Acceptable fields are.*"
 
                    ["extract" ["certname" "nothing" "nothing2"] ["~" "certname" ".*"]]
-                   #"Can't extract unknown 'events' fields: 'nothing', 'nothing2'.*Acceptable fields are.*")))
+                   #"Can't extract unknown 'events' fields 'nothing' and 'nothing2'.*Acceptable fields are.*")))
 
 (deftest-http-app invalid-queries
   [[version endpoint] endpoints
@@ -530,7 +529,7 @@
             (query-response method endpoint query)]
         (is (re-find msg body))
         (is (= status http/status-bad-request))
-        (is (= headers {"Content-Type" http/error-response-content-type}))))))
+        (are-error-response-headers headers)))))
 
 (def pg-versioned-invalid-regexps
   (omap/ordered-map
@@ -551,4 +550,4 @@
             (query-response method endpoint query)]
         (is (re-find msg body))
         (is (= status http/status-bad-request))
-        (is (= headers {"Content-Type" http/error-response-content-type}))))))
+        (are-error-response-headers headers)))))
