@@ -8,13 +8,15 @@
             [puppetlabs.puppetdb.testutils :refer [dotestseq]]
             [puppetlabs.puppetdb.testutils.db :refer [with-test-db]]
             [puppetlabs.puppetdb.testutils.http
-             :refer [deftest-http-app
+             :refer [are-error-response-headers
+                     deftest-http-app
                      ordered-query-result
                      query-response
                      query-result
                      vector-param
                      with-http-app]]))
 
+;; Queries issued at the root query endpoint
 (def endpoints [[:v4 "/v4"]])
 
 (deftest-http-app index-queries
@@ -26,15 +28,19 @@
         facts1  (assoc facts "fqdn" "host1")
         facts2  (assoc facts "fqdn" "host2")
         facts3  (assoc facts "fqdn" "host3")
+        facts4  (assoc facts "fqdn" "host4")
         cat1    (assoc catalog :certname "host1")
         cat2    (assoc catalog :certname "host2")
-        cat3    (assoc catalog :certname "host3")]
+        cat3    (assoc catalog :certname "host3")
+        cat4    (assoc catalog :certname "host4")]
     (scf-store/add-certname! "host1")
     (scf-store/add-certname! "host2")
     (scf-store/add-certname! "host3")
+    (scf-store/add-certname! "host4")
     (scf-store/replace-catalog! cat1 (now))
     (scf-store/replace-catalog! cat2 (now))
     (scf-store/replace-catalog! cat3 (now))
+    (scf-store/replace-catalog! cat4 (now))
     (scf-store/add-facts! {:certname "host1"
                            :values facts1
                            :timestamp (now)
@@ -53,23 +59,29 @@
                            :environment "DEV"
                            :producer_timestamp (now)
                            :producer "foo3"})
-    (scf-store/deactivate-node! "host3")
+    (scf-store/add-facts! {:certname "host4"
+                           :values facts4
+                           :timestamp (now)
+                           :environment "DEV"
+                           :producer_timestamp (now)
+                           :producer "foo4"})
+    (scf-store/deactivate-node! "host4")
 
     (testing "invalid from query"
       (let [{:keys [status body headers]} (query-response method endpoint ["from" "foobar"])]
         (is (re-find #"Invalid entity" body))
-        (is (= headers {"Content-Type" http/error-response-content-type}))
+        (are-error-response-headers headers)
         (is (= status http/status-bad-request)))
 
       ;; Ensure we parse anything that looks like AST/JSON as JSON not PQL
       (let [{:keys [status body headers]} (query-response method endpoint "[\"from\",\"foobar\"")]
         (is (= "Malformed JSON for query: [\"from\",\"foobar\"" body))
-        (is (= headers {"Content-Type" http/error-response-content-type}))
+        (are-error-response-headers headers)
         (is (= http/status-bad-request status)))
 
       (let [{:keys [status body headers]} (query-response method endpoint "foobar {}")]
         (is (re-find #"PQL parse error at line 1, column 1" body))
-        (is (= headers {"Content-Type" http/error-response-content-type}))
+        (are-error-response-headers headers)
         (is (= status http/status-bad-request))))
 
     (testing "pagination"
@@ -179,13 +191,42 @@
           (is (= results #{{:certname "host2"}})))))
 
     (testing "nodes"
-      (testing "query should return all nodes (including deactivated ones)"
+      (testing "query should return only active nodes"
         (doseq [query [["from" "nodes"]
-                       "nodes {}"]]
+                       ["from" "nodes" ["=" ["node" "active"] true]]
+                       ["from" "nodes" ["=" "node_state" "active"]]
+                       "nodes {}"
+                       "nodes { node_state = 'active' }"]]
           (is (= (set (mapv :certname (query-result method endpoint query {})))
                  #{"host1" "host2" "host3"}))))
 
-      (testing "broad regexp query should return all nodes"
+      (testing "query should return only inactive nodes when specified"
+        (doseq [query [["from" "nodes" ["or"
+                                        ["=" ["node" "active"] false]]]
+                       ["from" "nodes" ["=" "node_state" "inactive"]]
+                       "nodes { node_state = 'inactive' }"]]
+          (is (= (set (mapv :certname (query-result method endpoint query {})))
+                 #{"host4"}))))
+
+      (testing "query should return all nodes when specified"
+        (doseq [query [["from" "nodes" ["or"
+                                        ["=" ["node" "active"] true]
+                                        ["=" ["node" "active"] false]]]
+                       ["from" "nodes" ["=" "node_state" "any"]]
+                       "nodes { node_state = 'any' }"]]
+          (is (= (set (mapv :certname (query-result method endpoint query {})))
+                 #{"host1" "host2" "host3" "host4"}))))
+
+      (testing "query with 'and' and node_state = 'any' should work"
+        (doseq [query [["from" "nodes" ["or"
+                                        ["=" ["node" "active"] true]
+                                        ["=" ["node" "active"] false]]]
+                       ["from" "nodes" ["=" "node_state" "any"]]
+                       "nodes { certname ~ '^host\\d' and node_state = 'any' }"]]
+          (is (= (set (mapv :certname (query-result method endpoint query {})))
+                 #{"host1" "host2" "host3" "host4"}))))
+
+      (testing "broad regexp query should return all active nodes"
         (doseq [query [["from" "nodes" ["~" "certname" "^host"]]
                        "nodes { certname ~ '^host' }"]]
           (is (= (set (mapv :certname (query-result method endpoint query {})))
@@ -198,13 +239,11 @@
           (let [results (query-result method endpoint query)
                 result (first results)]
             (is (= host (:certname result)))
-            (if (= host "host3")
-              (is (:deactivated result))
-              (is (nil? (:deactivated result))))))))
+            (is (nil? (:deactivated result)))))))
 
     (testing "resources"
       (testing "query should return the resources just for that node"
-        (doseq [host ["host1" "host2"]
+        (doseq [host ["host1" "host2" "host3"]
                 query [["from" "resources" ["=" "certname" host]]
                        (format "resources { certname = '%s' }" host)]]
           (let [results (query-result method endpoint query)]
