@@ -6,6 +6,7 @@
             [puppetlabs.puppetdb.testutils :refer [paged-results] :as tu]
             [puppetlabs.puppetdb.testutils.http
              :refer [*app*
+                     are-error-response-headers
                      deftest-http-app
                      convert-response
                      ordered-query-result
@@ -46,7 +47,7 @@
                :catalog_environment :facts_environment :report_environment
                :latest_report_status :latest_report_hash :latest_report_noop
                :latest_report_noop_pending :cached_catalog_status
-               :latest_report_corrective_change} (keyset res))
+               :latest_report_corrective_change :latest_report_job_id} (keyset res))
           (str "Query was: " query))
       (is (= (set expected) (set (mapv :certname result)))
           (str "Query was: " query)))
@@ -87,6 +88,15 @@
     (testing "all nodes have last corrective_change status null"
       (is-query-result' ["null?" "latest_report_corrective_change" true] [db puppet web1 web2]))
 
+    (testing "query on last job_id returns an empty result"
+      (is-query-result' ["=" "latest_report_job_id" "1234567890"] []))
+
+    (testing "query on last job_id returns web1"
+      (is-query-result' ["=" "latest_report_job_id" "0987654321"] [web1]))
+
+    (testing "all nodes have last job_id null"
+      (is-query-result' ["null?" "latest_report_job_id" true] [db puppet web2]))
+
     (testing "regular expressions are supported for name"
       (is-query-result' ["~" "certname" "web\\d+.example.com"] [web1 web2])
       (is-query-result' ["~" "certname" "\\w+.example.com"] [db puppet web1 web2])
@@ -412,7 +422,7 @@
       (testing (str endpoint " query: " query " should fail with msg: " msg)
         (let [{:keys [status body headers]} (query-response method endpoint query)]
           (is (= status http/status-bad-request))
-          (is (= headers {"Content-Type" http/error-response-content-type}))
+          (are-error-response-headers headers)
           (is (re-find msg body)))))))
 
 (deftest-http-app query-with-pretty-printing
@@ -557,7 +567,11 @@
 
       (is-query-result' ["=" "report_timestamp" web1-report-ts] [web1])
       (is-query-result' [">" "report_timestamp" web1-report-ts] [db puppet])
-      (is-query-result' [">=" "report_timestamp" web1-report-ts] [web1 db puppet]))))
+      (is-query-result' [">=" "report_timestamp" web1-report-ts] [web1 db puppet])))
+
+  (let [res (->> ["<" "catalog_timestamp" "'2018-08-15 21:11:21 UTC'"]
+                 (query-response method endpoint))]
+    (is (= 400 (:status res)))))
 
 (def invalid-projection-queries
   (omap/ordered-map
@@ -566,7 +580,7 @@
      #"Can't extract unknown 'nodes' field 'nothing'.*Acceptable fields are.*"
 
      ["extract" ["certname" "nothing" "nothing2"] ["~" "certname" ".*"]]
-     #"Can't extract unknown 'nodes' fields: 'nothing', 'nothing2'.*Acceptable fields are.*"))
+     #"Can't extract unknown 'nodes' fields 'nothing' and 'nothing2'.*Acceptable fields are.*"))
 
 (deftest-http-app invalid-projections
   [[version endpoint] endpoints
@@ -576,7 +590,7 @@
     (let [{:keys [status body headers] :as result} (query-response method endpoint query)]
       (is (re-find msg body))
       (is (= status http/status-bad-request))
-      (is (= headers {"Content-Type" http/error-response-content-type})))))
+      (are-error-response-headers headers))))
 
 (def pg-versioned-invalid-regexps
   (omap/ordered-map
