@@ -8,6 +8,7 @@
             [puppetlabs.puppetdb.testutils.resources :refer [store-example-resources]]
             [puppetlabs.puppetdb.testutils.http
              :refer [*app*
+                     are-error-response-headers
                      deftest-http-app
                      query-response
                      ordered-query-result
@@ -24,7 +25,7 @@
 to the result of the form supplied to this method."
   [response body]
   (is (= http/status-ok (:status response)))
-  (is (= http/json-response-content-type (tu/content-type response)))
+  (is (http/json-utf8-ctype? (tu/content-type response)))
   (is (= body (if (:body response)
                 (set (json/parse-string (slurp (:body response)) true))
                 nil))))
@@ -84,8 +85,14 @@ to the result of the form supplied to this method."
 
     (testing "dot-style querying for regex resource parameters"
       (doseq [[query result] [[["~" "parameters.owner" "oot"] #{foo1 bar1}]
-                              [["~" "parameters.nested.foo" "ar"] #{foo1 bar1}]]]
-        (is (= (query-result (query-response method endpoint query)) result))))
+                              [["~" "parameters.owner" "^root"] #{foo1 bar1}]
+                              [["~" "parameters.owner" "oot$"] #{foo1 bar1}]
+                              [["~" "parameters.nested.foo" "ar"] #{foo1 bar1}]
+                              [["~" "parameters.nested.foo" "^bar$"] #{foo1 bar1}]
+                              [["~" "parameters.double_quote" "^foo\"bar$"] #{foo1 bar1}]
+                              [["~" "parameters.backslash" "^foo\\\\bar$"] #{foo1 bar1}]]]
+        (testing query
+          (is (= (query-result (query-response method endpoint query)) result)))))
 
     (testing "fact subqueries are supported"
       (let [{:keys [body status]}
@@ -101,12 +108,12 @@ to the result of the form supplied to this method."
         (is (= status http/status-ok))
         (is (= (set (json/parse-string (slurp body) true)) #{foo1})))
 
-      ;; Using the value of a fact as the title of a resource
-      (let [{:keys [body status]} (query-response method endpoint
-                                                ["in" "title" ["extract" "value" ["select_facts"
-                                                                                  ["=" "name" "message"]]]])]
-        (is (= status http/status-ok))
-        (is (= (set (json/parse-string (slurp body) true)) #{foo2 bar2}))))
+      (testing "using the value of a fact as the title of a resource"
+        (let [{:keys [body status]} (query-response method endpoint
+                                                    ["in" "title" ["extract" "value" ["select_facts"
+                                                                                      ["=" "name" "message"]]]])]
+          (is (= status http/status-ok))
+          (is (= (set (json/parse-string (slurp body) true)) #{foo2 bar2})))))
 
     (testing "resource subqueries are supported"
       ;; Fetch exported resources and their corresponding collected versions
@@ -124,7 +131,7 @@ to the result of the form supplied to this method."
       (let [response (query-response method endpoint ["="])
             body     (get response :body "null")]
         (is (= (:status response) http/status-bad-request))
-        (is (= (:headers response) {"Content-Type" http/error-response-content-type}))
+        (are-error-response-headers (:headers response))
         (is (re-find #"= requires exactly two arguments" body))))
 
     (testing "query with filter should exclude deactivated nodes"
@@ -193,18 +200,18 @@ to the result of the form supplied to this method."
       (let [query ["=" "sourceline" 22]
             response (query-response method endpoint query)]
         (is (= http/status-bad-request (:status response)))
-        (is (= (:headers response) {"Content-Type" http/error-response-content-type}))
-        (is (re-find #"'sourceline' is not a queryable object for resources, known queryable objects are" (:body response))))
+        (are-error-response-headers (:headers response))
+        (is (re-find #"'sourceline' is not a queryable object for resources. Known queryable objects are" (:body response))))
       (let [query ["~" "sourcefile" "foo"]
             response (query-response method endpoint query)]
         (is (= http/status-bad-request (:status response)))
-        (is (= (:headers response) {"Content-Type" http/error-response-content-type}))
-        (is (re-find #"'sourcefile' is not a queryable object for resources, known queryable objects are" (:body response))))
+        (are-error-response-headers (:headers response))
+        (is (re-find #"'sourcefile' is not a queryable object for resources. Known queryable objects are" (:body response))))
       (let [query ["=" "sourcefile" "/foo/bar"]
             response (query-response method endpoint query)]
         (is (= http/status-bad-request (:status response)))
-        (is (= (:headers response) {"Content-Type" http/error-response-content-type}))
-        (is (re-find #"'sourcefile' is not a queryable object for resources, known queryable objects are" (:body response)))))
+        (are-error-response-headers (:headers response))
+        (is (re-find #"'sourcefile' is not a queryable object for resources. Known queryable objects are" (:body response)))))
 
     (testing "query by file and line is supported"
       (let [query ["=" "file" "/foo/bar"]
@@ -350,7 +357,7 @@ to the result of the form supplied to this method."
                       #"Can't extract unknown 'resources' field 'nothing'.*Acceptable fields are.*"
 
                       ["extract" ["certname" "nothing" "nothing2"] ["~" "certname" ".*"]]
-                      #"Can't extract unknown 'resources' fields: 'nothing', 'nothing2'.*Acceptable fields are.*")))
+                      #"Can't extract unknown 'resources' fields 'nothing' and 'nothing2'.*Acceptable fields are.*")))
 
 (deftest-http-app invalid-queries
   [[version endpoint] endpoints
@@ -361,4 +368,4 @@ to the result of the form supplied to this method."
       (let [{:keys [status body headers] :as result} (query-response method endpoint query)]
         (is (re-find msg body))
         (is (= status http/status-bad-request))
-        (is (= headers {"Content-Type" http/error-response-content-type}))))))
+        (are-error-response-headers headers)))))
