@@ -345,6 +345,12 @@
       "certname > 4"
       [:condexpression [:field "certname"] ">" [:integer "4"]]
 
+      "certname != 4"
+      [:condexpression [:field "certname"] "!=" [:integer "4"]]
+
+      "certname !~ 4"
+      [:condexpression [:field "certname"] "!~" [:integer "4"]]
+
       "a in nodes [a] {}"
       [:condexpression [:field "a"] "in" [:from "nodes" [:extract [:field "a"]]]])
 
@@ -546,6 +552,7 @@
          "facts.operatingsystem.κόσμε"
          "facts.\"quoted field\".foo"
          "facts.\"field.with.dot\".foo"
+         "facts.\"field-with-dash\".foo"
          "trusted.authenticated"
          "parameters.😁"
          "latest_report?")
@@ -635,13 +642,14 @@
       "/asdf/"
       ""))
 
-  (testing "valuematch"
-    (are [in expected] (= (parse in :start :valuematch) expected)
+  (testing "literal"
+    (are [in expected] (= (parse in :start :literal) expected)
       "'asdf'" [[:sqstring "asdf"]]
       "1" [[:integer "1"]]
-      "1.1" [[:real "1" "." "1"]])
+      "1.1" [[:real "1" "." "1"]]
+      "[\"foo\", \"bar\"]" [[:array "[" [:dqstring "foo"] [:dqstring "bar"] "]"]])
 
-    (are [in] (insta/failure? (insta/parse parse in :start :valuematch))
+    (are [in] (insta/failure? (insta/parse parse in :start :literal))
       "asdf"
       ""))
 
