@@ -26,9 +26,10 @@
 
 (defn apply-migration-for-testing!
   [i]
-  (let [migration (migrations i)]
-    (migration)
-    (record-migration! i)))
+  (let [migration (migrations i)
+        result (migration)]
+    (record-migration! i)
+    result))
 
 (defn fast-forward-to-migration!
   [migration-number]
@@ -845,24 +846,6 @@
             (clojure.pprint/pprint unex))))
       (is (= expected diff)))))
 
-(deftest trgm-indexes-as-expected
-  ;; Assume the current *db* supports trgm
-  (clear-db-for-testing!)
-  (migrate! *db*)
-  (indexes! (:database *db*))
-  (let [idxs (:indexes (schema-info-map *db*))]
-    (is (= {:schema "public"
-            :table "facts"
-            :index "facts_value_string_trgm"
-            :index_keys ["value_string"]
-            :type "gin"
-            :unique? false
-            :functional? false
-            :is_partial false
-            :primary? false
-            :user "pdb_test"}
-           (get idxs ["facts" ["value_string"]])))))
-
 (deftest migration-60-fix-missing-edges-fk-constraint
   (jdbc/with-db-connection *db*
     (clear-db-for-testing!)
@@ -904,4 +887,151 @@
                                    :same nil}]}
                schema-diff))))))
 
+(deftest migration-64-rededuplicate-facts
+  (jdbc/with-db-connection *db*
+    (clear-db-for-testing!)
+    (fast-forward-to-migration! 63)
 
+    (jdbc/insert! :environments {:id 0 :environment "testing"})
+
+    (jdbc/insert! :certnames {:certname "a.com"})
+    (jdbc/insert! :certnames {:certname "b.com"})
+
+    (jdbc/insert! :factsets {:id 0
+                             :certname "a.com"
+                             :timestamp (to-timestamp (now))
+                             :producer_timestamp (to-timestamp (now))
+                             :environment_id 0
+                             :hash (sutils/munge-hash-for-storage "abcd1234")})
+
+    (jdbc/insert! :factsets {:id 1
+                             :certname "b.com"
+                             :timestamp (to-timestamp (now))
+                             :producer_timestamp (to-timestamp (now))
+                             :environment_id 0
+                             :hash (sutils/munge-hash-for-storage "1234abcd")})
+
+    (doseq [[id name] [[0 "string_fact"]
+                       [1 "int_fact"]
+                       [2 "float_fact"]
+                       [3 "bool_fact"]
+                       [4 "null_fact"]
+                       [5 "json_fact"]
+                       [6 "json_null_with_json_type"]
+                       [7 "json_null_with_null_type"]]]
+      (jdbc/insert! :fact_paths {:id id :depth 0 :name name :path name}))
+
+    (doseq [[fact-path-id value-type-id value_key value]
+            [[0 0 :value_string "string_value"]
+             [1 1 :value_integer 42]
+             [2 2 :value_float 4.2]
+             [3 3 :value_boolean false]
+             [4 4 :value_null nil]
+             [5 5 :value_json {:foo "bar"}]
+             ;; some databases have a fact value for both json null and sql null
+             [6 5 :value_json ::json-null]
+             [7 4 :value_null ::json-null]]
+            factset-id [0 1]]
+      (let [row {:factset_id factset-id
+                 :fact_path_id fact-path-id
+                 :value_type_id value-type-id
+                 ;; reduplicated facts code stored sql NULL for nil values in
+                 ;; this column, not json null
+                 :value (case value
+                          nil nil
+                          ::json-null (sutils/munge-jsonb-for-storage nil)
+                          (sutils/munge-jsonb-for-storage value))}]
+        (jdbc/insert! :facts (case value_key
+                               (:value_null :value_json) row
+                               (assoc row value_key value)))))
+
+    (is (= {::migrate/vacuum-analyze #{"facts" "fact_values" "fact_paths"}}
+           (apply-migration-for-testing! 64)))
+
+    (is (= 6 (:count (first (jdbc/query-to-vec "select count(*) from fact_values")))))
+    (is (= 16 (:count (first (jdbc/query-to-vec "select count(*) from facts")))))))
+
+(deftest migration-64-rededuplicate-facts-with-no-good-null-value
+  (jdbc/with-db-connection *db*
+    (clear-db-for-testing!)
+    (fast-forward-to-migration! 63)
+
+    (jdbc/insert! :environments {:id 0 :environment "testing"})
+    (jdbc/insert! :certnames {:certname "a.com"})
+    (jdbc/insert! :factsets {:id 0
+                             :certname "a.com"
+                             :timestamp (to-timestamp (now))
+                             :producer_timestamp (to-timestamp (now))
+                             :environment_id 0
+                             :hash (sutils/munge-hash-for-storage "abcd1234")})
+
+
+    (jdbc/insert! :fact_paths {:id 0
+                               :depth 0
+                               :name "json_null_with_json_type"
+                               :path "json_null_with_json_type"})
+
+    (jdbc/insert! :facts {:factset_id 0
+                          :fact_path_id 0
+                          :value_type_id 5
+                          :value (sutils/munge-jsonb-for-storage nil)})
+
+
+    (jdbc/insert! :fact_paths {:id 1
+                               :depth 0
+                               :name "json_null_with_null_type"
+                               :path "json_null_with_null"})
+
+    (jdbc/insert! :facts {:factset_id 1
+                          :fact_path_id 1
+                          :value_type_id 5
+                          :value (sutils/munge-jsonb-for-storage nil)})
+
+
+    (jdbc/insert! :fact_paths {:id 2
+                               :depth 0
+                               :name "sql_null_with_json_type"
+                               :path "sql_null_with_json_type"})
+
+    (jdbc/insert! :facts {:factset_id 2
+                          :fact_path_id 2
+                          :value_type_id 5
+                          :value (sutils/munge-jsonb-for-storage nil)})
+
+
+    (is (= {::migrate/vacuum-analyze #{"facts" "fact_values" "fact_paths"}}
+           (apply-migration-for-testing! 64)))
+
+    (is (= 1 (:count (first (jdbc/query-to-vec "select count(*) from fact_values")))))))
+
+(deftest migration-64-fixes-single-fact-with-sql-null-in-json-type
+  (jdbc/with-db-connection *db*
+    (clear-db-for-testing!)
+    (fast-forward-to-migration! 63)
+
+    (jdbc/insert! :environments {:id 0 :environment "testing"})
+    (jdbc/insert! :certnames {:certname "a.com"})
+    (jdbc/insert! :factsets {:id 0
+                             :certname "a.com"
+                             :timestamp (to-timestamp (now))
+                             :producer_timestamp (to-timestamp (now))
+                             :environment_id 0
+                             :hash (sutils/munge-hash-for-storage "abcd1234")})
+
+    (jdbc/insert! :fact_paths {:id 0
+                               :depth 0
+                               :name "sql_null_with_json_type"
+                               :path "sql_null_with_json_type"})
+
+    (jdbc/insert! :facts {:factset_id 0
+                          :fact_path_id 0
+                          :value_type_id 5
+                          :value nil})
+
+
+    (is (= {::migrate/vacuum-analyze #{"facts" "fact_values" "fact_paths"}}
+           (apply-migration-for-testing! 64)))
+
+    (is (= [{:value_type_id 4
+              :value nil}]
+           (jdbc/query-to-vec "select value_type_id, value from fact_values")))))
